import inspect
import traceback

from django.test.client import (
    Client,
)

from behave_bo.loggers import (
    tests_logger,
)


class BehaveBOClient(Client):

    def __init__(self):
        super().__init__()
        self.non_200_requests = list()
        self.code_parts = ['AppRequest(', '.get_raw_data(']

    def find_code_context(self):
        stack = inspect.stack()
        file_name = None
        break_flag = False
        for frame in stack:
            for code in frame.code_context:
                if any(part in code for part in self.code_parts):
                    file_name = f"{frame.filename}:{frame.lineno}"
                    break_flag = True
                    break
            if break_flag:
                break

        if file_name is None:
            for line in traceback.format_stack():
                print(line.strip())

        return file_name

    def request(self, **request):
        response = super().request(**request)

        if response.status_code != 200:
            file_name = self.find_code_context()
            self.non_200_requests.append((file_name, response.request["PATH_INFO"], response.status_code))

        return response

    def print_non_200_requests(self) -> None:
        """Выводит в консоль список запросов, которые завершились с ошибкой"""
        if self.non_200_requests:
            tests_logger.error('Список запросов к приложению, завершившихся с ошибкой:')
            for file_name, request_path, status_code in self.non_200_requests:
                tests_logger.error(f'{file_name} - {request_path} - {status_code}')
        self.non_200_requests = list()



import json
import re


class ContextExtractor:
    """
    Утилита для преобразования контекста тестирования, ui-компонента или
    содержимого ответа на запрос в нужный для автотестов вид.
    """
    regex_patterns = []

    def extract_flat_dict_from_component(self, component, is_child_item=False, data=None):
        """Преобразовывает Ext-контейнер/поле в плоский словарь с ключами-значениями из вложенных полей.

        Args:
            component: Объект Ext-контейнера/поля.
            is_child_item: Признак вызова для дочернего компонента (в рекурсии).
            data: Словарь пробрасываемый при рекурсивном вызове.
        Returns:
            Плоский словарь содержащий значения полей.
        """
        if not data:
            data = {}

        return data

    def prepare_component_json(self, component):
        """Преобразует компонент в словарь и затем в json-строку.

        Args:
            component: Объект Ext-компонента.

        Returns:
            JSON-строка.
        """
        return json.dumps(
            self.extract_flat_dict_from_component(component),
            default=str,
        )

    def extract_data_with_regex(self, response_content) -> dict:
        """Извлекает из js-ответа данные с помощью регулярных выражений.

        Args:
            response_content: Контент ответа на запрос

        Returns:
            Словарь с результатами поиска.
        """
        response_content = response_content.decode()
        result = {}

        # В текущей реализации порядок добавления re паттернов имеет значение.
        for pattern in self.regex_patterns:
            result = pattern.search(response_content)
            if result:
                result = result.groupdict()
                for key, value in result.items():
                    try:
                        result[key] = json.loads(value)
                    except ValueError:
                        pass
                break

        return result

    def parse_download_file_name(self, content, download_url) -> str:
        """Парсит наименование скачиваемого файла.

        Args:
            content: Контент ответа на запрос.
            download_url: Путь скачивания файла

        Returns:
            Наименование файла.

        Raises:
            Exception: если параметры файла не найдены
        """
        match = re.findall(
            rf"(?<={download_url}).*?(?=\'|\"|$)",
            content
        )

        if match:
            file_name = match[0]
            if bool(re.search('[\\\\u]', file_name)):
                file_name = file_name.encode().decode('unicode-escape')
        else:
            raise Exception('Имя файла не найдено')

        return file_name



"""
Knowledge base of all built-in formatters.
"""

from behave_bo.formatter import (
    _registry,
)


# -----------------------------------------------------------------------------
# DATA:
# -----------------------------------------------------------------------------
# SCHEMA: formatter.name, formatter.class(_name)
_BUILTIN_FORMATS = [
    ("plain", "behave_bo.formatter.plain:PlainFormatter"),
    ("pretty", "behave_bo.formatter.pretty:PrettyFormatter"),
    ("json", "behave_bo.formatter.json:JSONFormatter"),
    ("json.pretty", "behave_bo.formatter.json:PrettyJSONFormatter"),
    ("null", "behave_bo.formatter.null:NullFormatter"),
    ("progress", "behave_bo.formatter.progress:ScenarioProgressFormatter"),
    ("progress2", "behave_bo.formatter.progress:StepProgressFormatter"),
    ("progress3", "behave_bo.formatter.progress:ScenarioStepProgressFormatter"),
    ("rerun", "behave_bo.formatter.rerun:RerunFormatter"),
    ("tags", "behave_bo.formatter.tags:TagsFormatter"),
    ("tags.location", "behave_bo.formatter.tags:TagsLocationFormatter"),
    ("steps", "behave_bo.formatter.steps:StepsFormatter"),
    ("steps.doc", "behave_bo.formatter.steps:StepsDocFormatter"),
    ("steps.catalog", "behave_bo.formatter.steps:StepsCatalogFormatter"),
    ("steps.usage", "behave_bo.formatter.steps:StepsUsageFormatter"),
    ("steps.remove_unused", "behave_bo.formatter.steps:UnusedStepsRemoveFormatter"),
    ("sphinx.steps", "behave_bo.formatter.sphinx_steps:SphinxStepsFormatter"),
]

# -----------------------------------------------------------------------------
# FUNCTIONS:
# -----------------------------------------------------------------------------
def setup_formatters():
    """Register all built-in formatters (lazy-loaded)."""
    _registry.register_formats(_BUILTIN_FORMATS)



"""
This module provides `JSON`_ formatters for :mod:`behave_bo`:

* json: Generates compact JSON output
* json.pretty: Generates readable JSON output

.. _JSON: http://json.org
"""

import base64
import json

from behave_bo.formatter.base import (
    Formatter,
)
from behave_bo.model_core import (
    Status,
)


# -----------------------------------------------------------------------------
# CLASS: JSONFormatter
# -----------------------------------------------------------------------------
class JSONFormatter(Formatter):
    name = "json"
    description = "JSON dump of test run"
    dumps_kwargs = {}
    split_text_into_lines = True   # EXPERIMENT for better readability.

    json_number_types = (int,) + (float,)
    json_scalar_types = json_number_types + (str, bool, type(None))

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)
        # -- ENSURE: Output stream is open.
        self.stream = self.open()
        self.feature_count = 0
        self.current_feature = None
        self.current_feature_data = None
        self.current_scenario = None
        self._step_index = 0

    def reset(self):
        self.current_feature = None
        self.current_feature_data = None
        self.current_scenario = None
        self._step_index = 0

    # -- FORMATTER API:
    def uri(self, uri):
        pass

    def feature(self, feature):
        self.reset()
        self.current_feature = feature
        self.current_feature_data = {
            "keyword": feature.keyword,
            "name": feature.name,
            "tags": list(feature.tags),
            "location": str(feature.location),
            "status": None,     # Not known before feature run.
        }
        element = self.current_feature_data
        if feature.description:
            element["description"] = feature.description

    def background(self, background):
        element = self.add_feature_element({
            "type": "background",
            "keyword": background.keyword,
            "name": background.name,
            "location": str(background.location),
            "steps": [],
        })
        if background.name:
            element["name"] = background.name
        self._step_index = 0

        # -- ADD BACKGROUND STEPS: Support *.feature file regeneration.
        for step_ in background.steps:
            self.step(step_)

    def scenario(self, scenario):
        self.finish_current_scenario()
        self.current_scenario = scenario

        element = self.add_feature_element({
            "type": "scenario",
            "keyword": scenario.keyword,
            "name": scenario.name,
            "tags": scenario.tags,
            "location": str(scenario.location),
            "steps": [],
            "status": None,
        })
        if scenario.description:
            element["description"] = scenario.description
        self._step_index = 0

    @classmethod
    def make_table(cls, table):
        table_data = {
            "headings": table.headings,
            "rows": [list(row) for row in table.rows]
        }
        return table_data

    def step(self, step):
        s = {
            "keyword": step.keyword,
            "step_type": step.step_type,
            "name": step.name,
            "location": str(step.location),
        }

        if step.text:
            text = step.text
            if self.split_text_into_lines and "\n" in text:
                text = text.splitlines()
            s["text"] = text
        if step.table:
            s["table"] = self.make_table(step.table)
        element = self.current_feature_element
        element["steps"].append(s)

    def match(self, match):
        args = []
        for argument in match.arguments:
            argument_value = argument.value
            if not isinstance(argument_value, self.json_scalar_types):
                # -- OOPS: Avoid invalid JSON format w/ custom types.
                # Use raw string (original) instead.
                argument_value = argument.original
            assert isinstance(argument_value, self.json_scalar_types)
            arg = {
                "value": argument_value,
            }
            if argument.name:
                arg["name"] = argument.name
            if argument.original != argument_value:
                # -- REDUNDANT DATA COMPRESSION: Suppress for strings.
                arg["original"] = argument.original
            args.append(arg)

        match_data = {
            "location": str(match.location) or "",
            "arguments": args,
        }
        if match.location:
            # -- NOTE: match.location=None occurs for undefined steps.
            steps = self.current_feature_element["steps"]
            steps[self._step_index]["match"] = match_data

    def result(self, step):
        steps = self.current_feature_element["steps"]
        steps[self._step_index]["result"] = {
            "status": step.status.name,
            "duration": step.duration,
        }
        if step.error_message and step.status == Status.failed:
            # -- OPTIONAL: Provided for failed steps.
            error_message = step.error_message
            if self.split_text_into_lines and "\n" in error_message:
                error_message = error_message.splitlines()
            result_element = steps[self._step_index]["result"]
            result_element["error_message"] = error_message
        self._step_index += 1

    def embedding(self, mime_type, data):
        step = self.current_feature_element["steps"][-1]
        step["embeddings"].append({
            "mime_type": mime_type,
            "data": base64.b64encode(data).replace("\n", ""),
        })

    def eof(self):
        """
        End of feature
        """
        if not self.current_feature_data:
            return

        # -- NORMAL CASE: Write collected data of current feature.
        self.finish_current_scenario()
        self.update_status_data()

        if self.feature_count == 0:
            # -- FIRST FEATURE:
            self.write_json_header()
        else:
            # -- NEXT FEATURE:
            self.write_json_feature_separator()

        self.write_json_feature(self.current_feature_data)
        self.reset()
        self.feature_count += 1

    def close(self):
        if self.feature_count == 0:
            # -- FIRST FEATURE: Corner case when no features are provided.
            self.write_json_header()
        self.write_json_footer()
        self.close_stream()

    # -- JSON-DATA COLLECTION:
    def add_feature_element(self, element):
        assert self.current_feature_data is not None
        if "elements" not in self.current_feature_data:
            self.current_feature_data["elements"] = []
        self.current_feature_data["elements"].append(element)
        return element

    @property
    def current_feature_element(self):
        assert self.current_feature_data is not None
        return self.current_feature_data["elements"][-1]

    def update_status_data(self):
        assert self.current_feature
        assert self.current_feature_data
        self.current_feature_data["status"] = self.current_feature.status.name

    def finish_current_scenario(self):
        if self.current_scenario:
            status_name = self.current_scenario.status.name
            self.current_feature_element["status"] = status_name

    # -- JSON-WRITER:
    def write_json_header(self):
        self.stream.write("[\n")

    def write_json_footer(self):
        self.stream.write("\n]\n")

    def write_json_feature(self, feature_data):
        self.stream.write(json.dumps(feature_data, **self.dumps_kwargs))
        self.stream.flush()

    def write_json_feature_separator(self):
        self.stream.write(",\n\n")


# -----------------------------------------------------------------------------
# CLASS: PrettyJSONFormatter
# -----------------------------------------------------------------------------
class PrettyJSONFormatter(JSONFormatter):
    """
    Provides readable/comparable textual JSON output.
    """
    name = "json.pretty"
    description = "JSON dump of test run (human readable)"
    dumps_kwargs = {"indent": 2, "sort_keys": True}



"""
Functionality to support user-specific configuration data (userdata).
"""

from behave_bo._types import (
    Unknown,
)


# -----------------------------------------------------------------------------
# FUNCTIONS:
# -----------------------------------------------------------------------------
def parse_bool(text):
    """Parses a boolean text and converts it into boolean value (if possible).
    Supported truth string values:

      * true:   "true", "yes", "on", "1"
      * false:  "false", "no", "off", "0"

    :raises: ValueError, if text is invalid
    """
    from distutils.util import (
        strtobool,
    )
    return bool(strtobool(text))


def parse_user_define(text):
    """Parse "{name}={value}" text and return parts as tuple.
    Used for command-line definitions, like "... -D name=value".

    SUPPORTED SCHEMA:

      * "{name}={value}"
      * "{name}"                (boolean flag; value="true")
      * '"{name}={value}"'      (double-quoted name-value pair)
      * "'{name}={value}'"      (single-quoted name-value pair)
      * '{name}="{value}"'      (double-quoted value)
      * "{name}='{value}'"      (single-quoted value)
      * "  {name} = {value}  "  (whitespace padded)

    .. note:: Leading/trailing Quotes are stripped.

    :param text:    Text to parse (as string).
    :return: (name, value) pair as tuple.
    """
    text = text.strip()
    if "=" in text:
        text = unqote(text)
        name, value = text.split("=", 1)
        name = name.strip()
        value = unqote(value.strip())
    else:
        # -- ASSUMPTION: Boolean definition (as flag)
        name = text
        value = "true"
    return (name, value)


def unqote(text):
    """Strip pair of leading and trailing quotes from text."""
    # -- QUOTED: Strip single-quote or double-quote pair.
    if ((text.startswith('"') and text.endswith('"')) or
            (text.startswith("'") and text.endswith("'"))):
        text = text[1:-1]
    return text

# -----------------------------------------------------------------------------
# CLASSES:
# -----------------------------------------------------------------------------
class UserData(dict):
    """Dictionary-like user-data with some additional features:

      * type-converter methods, similar to configparser.ConfigParser.getint()

    """

    def getas(self, convert, name, default=None, valuetype=None):
        """Converts the value of user-data parameter from a string into a
        specific value type.

        :param convert: Converter function to use (string to value-type).
        :param name:    Variable name to use.
        :param default: Default value, used if parameter is not found.
        :param valuetype: Value type(s), needed if convert != valuetype()
        :return: Converted textual value (type: valuetype)
        :return: Default value, if parameter is unknown.
        :raises ValueError: If type conversion fails.
        """
        if valuetype is None:
            # -- ASSUME: Converter function is the type constructor.
            valuetype = convert

        value = self.get(name, Unknown)
        if value is Unknown:
            return default
        elif isinstance(value, valuetype):
            # -- PRESERVE: Pre-converted value if type matches.
            return value
        else:
            # -- CASE: Textual value (expected)
            # Raise ValueError if parse/conversion fails.
            assert callable(convert)
            return convert(value)

    def getint(self, name, default=0):
        """Convert parameter value (as string) into a integer value.
        :return: Parameter value as integer number (on success).
        :raises: ValueError, if type conversion fails.
        """
        return self.getas(int, name, default)

    def getfloat(self, name, default=0.0):
        """Convert parameter value (as string) into a float value.
        :return: Parameter value as float number (on success).
        :raises: ValueError, if type conversion fails.
        """
        return self.getas(float, name, default)

    def getbool(self, name, default=False):
        """Converts user-data string-value into boolean value (if possible).
        Supported truth string values:

          * true:   "true", "yes", "on", "1"
          * false:  "false", "no", "off", "0"

        :param name: Parameter name (as string).
        :param default: Default value, if parameter is unknown (=False).
        :return: Boolean value of parameter
        :raises: ValueError, if type conversion fails.
        """
        return self.getas(parse_bool, name, default, valuetype=bool)

    @classmethod
    def make(cls, data):
        if data is None:
            data = cls()
        elif not isinstance(data, cls):
            data = cls(data)
        return data


class UserDataNamespace:
    """Provides a light-weight dictview to the user data that allows you
    to access all params in a namespace, that use "{namespace}.*" names.

    .. code-block:: python

        my_config = UserDataNamespace("my.config", userdata)
        value1 = my_config.getint("value1")  # USE: my.config.value1
        value2 = my_config.get("value2")     # USE: my.config.value2
    """

    def __init__(self, namespace, data=None):
        self.namespace = namespace or ""
        self.data = UserData.make(data)

    @staticmethod
    def make_scoped(namespace, name):
        """Creates a scoped-name from its parts."""
        if not namespace:   # noqa
            return name
        return f"{namespace}.{name}"

    # -- DICT-LIKE:
    def get(self, name, default=None):
        scoped_name = self.make_scoped(self.namespace, name)
        return self.data.get(scoped_name, default)

    def getas(self, convert, name, default=None, valuetype=None):
        scoped_name = self.make_scoped(self.namespace, name)
        return self.data.getas(convert, scoped_name, default=default,
                               valuetype=valuetype)

    def getint(self, name, default=0):
        scoped_name = self.make_scoped(self.namespace, name)
        return self.data.getint(scoped_name, default=default)

    def getfloat(self, name, default=0.0):
        scoped_name = self.make_scoped(self.namespace, name)
        return self.data.getfloat(scoped_name, default=default)

    def getbool(self, name, default=False):
        scoped_name = self.make_scoped(self.namespace, name)
        return self.data.getbool(scoped_name, default=default)

    def __contains__(self, name):
        scoped_name = self.make_scoped(self.namespace, name)
        return scoped_name in self.data

    def __getitem__(self, name):
        scoped_name = self.make_scoped(self.namespace, name)
        return self.data[scoped_name]

    def __setitem__(self, name, value):
        scoped_name = self.make_scoped(self.namespace, name)
        self.data[scoped_name] = value

    def __len__(self):
        return len(self.scoped_keys())

    def scoped_keys(self):
        if not self.namespace:  # noqa
            return self.data.keys()
        prefix = "%s." % self.namespace
        return [key for key in self.data.keys() if key.startswith(prefix)]

    def keys(self):
        prefix = "%s." % self.namespace
        for scoped_name in self.scoped_keys():
            name = scoped_name.replace(prefix, "", 1)
            yield name

    def values(self):
        for scoped_name in self.scoped_keys():
            yield self.data[scoped_name]

    def items(self):
        for name in self.keys():
            scoped_name = self.make_scoped(self.namespace, name)
            value = self.data[scoped_name]
            yield (name, value)



import json
import os
import sys
from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Iterable,
)

from django.conf import (
    settings,
)

from behave_bo.configuration import (
    Configuration,
)
from behave_bo.formatter.base import (
    StreamOpener,
)
from behave_bo.formatter.progress import (
    ScenarioStepProgressFormatter,
)
from behave_bo.model import (
    ScenarioOutline,
)
from behave_bo.model_core import (
    Status,
)
from behave_bo.reporter.base import (
    Reporter,
)
from behave_bo.tag_expression import (
    TagExpression,
)


class FalsePositiveTestsReport:
    """
    Класс для обработки информации по ложным падениям тестов.
    """

    def __init__(self, build_url: str = ''):
        self.is_opened = False

        self.build_url = build_url
        self.date_format = '%Y-%m-%d'
        self.storage_period = timedelta(days=60)

        self.report_file_path = settings.WEB_BB_BEHAVE__FALSE_POSITIVE_TESTS_INFO
        self.report = {}

    def open(self, actualize_file: bool = False):
        """Открывает файл с отчетом.

        Args:
            actualize_file: Выполнить актуализацию содержимого файла в соответствии с периодом хранения.
        """
        if not self.is_opened:
            if os.path.exists(self.report_file_path):
                with open(self.report_file_path) as f:
                    self.report = json.load(f)

            if self.report and actualize_file:
                self._actualize()
                self._save()

            self.is_opened = True

    def update_and_save(self, tags: Iterable):
        """Обновляет информацию о ложных падениях и сохраняет в файл.

        Args:
            tags: Список тегов.
        """
        if tags:
            self._update(tags)
            self._save()

    def _update(self, tags: Iterable):
        """Обновляет информацию о ложных падениях.

        Args:
            tags: Список тегов.
        """
        current_date_string = datetime.today().strftime(self.date_format)

        for tag in tags:
            if tag in self.report:
                self.report[tag]['count'] += 1
                self.report[tag]['failures'].setdefault(
                    current_date_string,
                    []
                ).append(self.build_url)
            else:
                self.report[tag] = {
                    'count': 1,
                    'failures': {
                        current_date_string: [self.build_url],
                    },
                }

    def _is_storage_expired(self, date_string: str) -> bool:
        """Проверяет, истек ли срок хранения информации по переданной дате.

        Args:
            date_string: Строковое представление даты заданного формата.

        Returns:
            Результат проверки.
        """

        return datetime.today() - datetime.strptime(date_string, self.date_format) > self.storage_period

    def _actualize(self):
        """Актуализирует информацию в соответствии с периодом хранения.

        Удаляет информацию по тесту, если дата последнего падения больше заданного срока хранения.
        Удаляет ссылки на сборки по тесту, если дата падения больше заданного срока хранения.
        """
        expired_tags = set()
        for tag, info in self.report.items():
            failure_dates = sorted(info['failures'])

            if self._is_storage_expired(failure_dates[-1]):
                expired_tags.add(tag)
            else:
                for date_string in failure_dates:
                    if self._is_storage_expired(date_string):
                        info['failures'].pop(date_string)
                    else:
                        break

                info['count'] = sum([len(list_of_urls) for list_of_urls in info['failures'].values()])

        for tag in expired_tags:
            self.report.pop(tag)

    def _save(self):
        """Сохраняет информацию о ложных падениях в файл."""
        try:
            with open(self.report_file_path, 'w') as f:
                f.writelines(json.dumps(self.report))
        except FileNotFoundError as e:
            print(f'Не удалось сохранить информацию о ложных падениях в файл: {e}')


class FailedTagsReporter(Reporter):
    """
    Репортер выводящий в лог список тегов сценариев упавших автотестов разделённых запятой.
    """

    def __init__(self, config: Configuration):
        super().__init__(config)
        stream = getattr(sys, "stdout", sys.stderr)
        self.stream = StreamOpener.ensure_stream_with_encoder(stream)
        self.failed_tags = set()
        self._main_run_failed_tags = set()
        self.is_in_rerun_mode = False
        self.false_positive_tests_amount = None

        self.false_positive_tests_report = FalsePositiveTestsReport(
            build_url=config.userdata.get('build_url', ''),
        )

    @property
    def main_run_failed_tags(self):
        """Теги упавших тестов основного прогона."""

        return self._main_run_failed_tags if self.is_in_rerun_mode else self.failed_tags

    @property
    def false_positive_tags(self) -> set:
        """Теги тестов с ложными падениями."""

        return self._main_run_failed_tags.difference(self.failed_tags) if self.is_in_rerun_mode else set()

    def rerun_required(self) -> bool:
        """Возвращает результат проверки условий перезапуска."""

        return (
            self.config.rerun_if_failed and
            self.failed_tags and
            not self.is_in_rerun_mode and
            len(self.failed_tags) <= self.config.rerun_if_failed
        )

    def rerun_up(self):
        """Выполняет настройки перед запуском упавших сценариев."""
        self.config.rerun = True
        self.config.tags = TagExpression([','.join(self.failed_tags)])
        self.config.reporters = [self]
        self.config.format = [ScenarioStepProgressFormatter.name]

        self._main_run_failed_tags = self.failed_tags
        self.failed_tags = set()
        self.is_in_rerun_mode = True
        self.false_positive_tests_report.open(actualize_file=True)

    def log_failed_tags(self):
        """Выводит информацию об упавших тестах в виде списка тегов."""
        self.stream.write(f"\nFailing scenarios tags: {','.join(self.main_run_failed_tags)}\n")
        if self.is_in_rerun_mode:
            self.false_positive_tests_amount = len(self.false_positive_tags) if self.failed_tags else 'All'
            self.stream.write(f"{self.false_positive_tests_amount} false-positive: {','.join(self.false_positive_tags)}\n")

            if self.failed_tags:
                self.stream.write(f"{len(self.failed_tags)} actual failed: {','.join(self.failed_tags)}\n")

        self.stream.write("\n")

    def feature(self, feature):
        for scenario in feature:
            if isinstance(scenario, ScenarioOutline):
                self.process_scenario_outline(scenario)
            else:
                self.process_scenario(scenario)

    def end(self):
        if self.failed_tags or self.is_in_rerun_mode:
            self.log_failed_tags()
            if self.false_positive_tags:
                self.false_positive_tests_report.update_and_save(self.false_positive_tags)

    def process_scenario(self, scenario):
        if scenario.status == Status.failed:
            self.failed_tags.add(f'{scenario.main_tag}')

    def process_scenario_outline(self, scenario_outline):
        for scenario in scenario_outline.scenarios:
            self.process_scenario(scenario)

    def check_all_false_positive(self):
        return self.false_positive_tests_amount == 'All'


class FailedTagsParallelReporter(FailedTagsReporter):
    """
    Репортер выводящий в лог список тегов сценариев упавших автотестов разделённых запятой адаптированный
    для запуска в параллельных процессах.
    """
    process_kwarg_name = 'failed_tags_proxy_list'
    failed_tags_proxy_list = None

    def init_process_variable(self, manager):
        """Инициализирует атрибут для хранения списка тегов.

        Args:
            manager: multiprocessing.Manager
        """
        self.failed_tags_proxy_list = manager.list()

        return self.failed_tags_proxy_list

    def end(self):
        if self.failed_tags:
            self.failed_tags_proxy_list.append(self.failed_tags)

    def after_parallel_run(self):
        if self.failed_tags_proxy_list or self.is_in_rerun_mode:
            self.failed_tags.update(*self.failed_tags_proxy_list)
            self.log_failed_tags()
            if self.false_positive_tags:
                self.false_positive_tests_report.update_and_save(self.false_positive_tags)



import argparse
import configparser
import logging
import os
import re
import shlex
import sys

from behave_bo._types import (
    Unknown,
)
from behave_bo.formatter import (
    _registry as _format_registry,
)
from behave_bo.formatter.base import (
    StreamOpener,
)
from behave_bo.model import (
    ScenarioOutline,
)
from behave_bo.model_core import (
    FileLocation,
)
from behave_bo.reporter.junit import (
    JUnitReporter,
)
from behave_bo.reporter.summary import (
    SummaryReporter,
)
from behave_bo.tag_expression import (
    TagExpression,
)
from behave_bo.textutil import (
    select_best_encoding,
    to_texts,
)
from behave_bo.userdata import (
    UserData,
    parse_user_define,
)


ConfigParser = configparser.ConfigParser


# -----------------------------------------------------------------------------
# CONFIGURATION DATA TYPES:
# -----------------------------------------------------------------------------
class LogLevel:
    names = [
        "NOTSET", "CRITICAL", "FATAL", "ERROR",
        "WARNING", "WARN", "INFO", "DEBUG",
    ]

    @staticmethod
    def parse(levelname, unknown_level=None):
        """
        Convert levelname into a numeric log level.

        :param levelname: Logging levelname (as string)
        :param unknown_level: Used if levelname is unknown (optional).
        :return: Numeric log-level or unknown_level, if levelname is unknown.
        """
        return getattr(logging, levelname.upper(), unknown_level)

    @classmethod
    def parse_type(cls, levelname):
        level = cls.parse(levelname, Unknown)
        if level is Unknown:
            message = "%s is unknown, use: %s" % \
                      (levelname, ", ".join(cls.names[1:]))
            raise argparse.ArgumentTypeError(message)
        return level

    @staticmethod
    def to_string(level):
        return logging.getLevelName(level)


class ConfigError(Exception):
    pass


# -----------------------------------------------------------------------------
# CONFIGURATION SCHEMA:
# -----------------------------------------------------------------------------
options = [
    (("-c", "--no-color"),
     dict(action="store_false", dest="color",
          help="Disable the use of ANSI color escapes.")),

    (("--color",),
     dict(action="store_true", dest="color",
          help="""Use ANSI color escapes. This is the default
                  behaviour. This switch is used to override a
                  configuration file setting.""")),

    (("-d", "--dry-run"),
     dict(action="store_true",
          help="Invokes formatters without executing the steps.")),

    (("-D", "--define"),
     dict(dest="userdata_defines", type=parse_user_define, action="append",
          metavar="NAME=VALUE",
          help="""Define user-specific data for the config.userdata dictionary.
                  Example: -D foo=bar to store it in config.userdata["foo"].""")),

    (("-e", "--exclude"),
     dict(metavar="PATTERN", dest="exclude_re",
          help="""Don't run feature files matching regular expression
                  PATTERN.""")),

    (("-i", "--include"),
     dict(metavar="PATTERN", dest="include_re",
          help="Only run feature files matching regular expression PATTERN.")),

    (("--no-junit",),
     dict(action="store_false", dest="junit",
          help="Don't output JUnit-compatible reports.")),

    (("--junit",),
     dict(action="store_true",
          help="""Output JUnit-compatible reports.
                  When junit is enabled, all stdout and stderr
                  will be redirected and dumped to the junit report,
                  regardless of the "--capture" and "--no-capture" options.
                  """)),

    (("--junit-directory",),
     dict(metavar="PATH", dest="junit_directory",
          default="reports",
          help="""ABSOLUTE PATH to directory in which to store JUnit reports.""")),

    ((),  # -- CONFIGFILE only
     dict(dest="default_format",
          help="Specify default formatter (default: pretty).")),


    (("-f", "--format"),
     dict(action="append",
          help="""Specify a formatter. If none is specified the default
                  formatter is used. Pass "--format help" to get a
                  list of available formatters.""")),

    (("--steps-catalog",),
     dict(action="store_true", dest="steps_catalog",
          help="""Show a catalog of all available step definitions.
                  SAME AS: --format=steps.catalog --dry-run --no-summary -q""")),

    ((),  # -- CONFIGFILE only
     dict(dest="scenario_outline_annotation_schema",
          help="""Specify name annotation schema for scenario outline
                  (default="{name} -- @{row.id} {examples.name}").""")),

    (("-k", "--no-skipped"),
     dict(action="store_false", dest="show_skipped",
          help="Don't print skipped steps (due to tags).")),

    (("--show-skipped",),
     dict(action="store_true",
          help="""Print skipped steps.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-snippets",),
     dict(action="store_false", dest="show_snippets",
          help="Don't print snippets for unimplemented steps.")),
    (("--snippets",),
     dict(action="store_true", dest="show_snippets",
          help="""Print snippets for unimplemented steps.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("-m", "--no-multiline"),
     dict(action="store_false", dest="show_multiline",
          help="""Don't print multiline strings and tables under
                  steps.""")),

    (("--multiline", ),
     dict(action="store_true", dest="show_multiline",
          help="""Print multiline strings and tables under steps.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("-n", "--name"),
     dict(action="append",
          help="""Only execute the feature elements which match part
                  of the given name. If this option is given more
                  than once, it will match against all the given
                  names.""")),

    (("--no-capture",),
     dict(action="store_false", dest="stdout_capture",
          help="""Don't capture stdout (any stdout output will be
                  printed immediately.)""")),

    (("--capture",),
     dict(action="store_true", dest="stdout_capture",
          help="""Capture stdout (any stdout output will be
                  printed if there is a failure.)
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-capture-stderr",),
     dict(action="store_false", dest="stderr_capture",
          help="""Don't capture stderr (any stderr output will be
                  printed immediately.)""")),

    (("--capture-stderr",),
     dict(action="store_true", dest="stderr_capture",
          help="""Capture stderr (any stderr output will be
                  printed if there is a failure.)
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--no-logcapture",),
     dict(action="store_false", dest="log_capture",
          help="""Don't capture logging. Logging configuration will
                  be left intact.""")),

    (("--logcapture",),
     dict(action="store_true", dest="log_capture",
          help="""Capture logging. All logging during a step will be captured
                  and displayed in the event of a failure.
                  This is the default behaviour. This switch is used to
                  override a configuration file setting.""")),

    (("--logging-level",),
     dict(type=LogLevel.parse_type,
          help="""Specify a level to capture logging at. The default
                  is INFO - capturing everything.""")),

    (("--logging-format",),
     dict(help="""Specify custom format to print statements. Uses the
                  same format as used by standard logging handlers. The
                  default is "%%(levelname)s:%%(name)s:%%(message)s".""")),

    (("--logging-datefmt",),
     dict(help="""Specify custom date/time format to print
                  statements.
                  Uses the same format as used by standard logging
                  handlers.""")),

    (("--logging-filter",),
     dict(help="""Specify which statements to filter in/out. By default,
                  everything is captured. If the output is too verbose, use
                  this option to filter out needless output.
                  Example: --logging-filter=foo will capture statements issued
                  ONLY to foo or foo.what.ever.sub but not foobar or other
                  logger. Specify multiple loggers with comma:
                  filter=foo,bar,baz.
                  If any logger name is prefixed with a minus, eg filter=-foo,
                  it will be excluded rather than included.""",
          config_help="""Specify which statements to filter in/out. By default,
                         everything is captured. If the output is too verbose,
                         use this option to filter out needless output.
                         Example: ``logging_filter = foo`` will capture
                         statements issued ONLY to "foo" or "foo.what.ever.sub"
                         but not "foobar" or other logger. Specify multiple
                         loggers with comma: ``logging_filter = foo,bar,baz``.
                         If any logger name is prefixed with a minus, eg
                         ``logging_filter = -foo``, it will be excluded rather
                         than included.""")),

    (("--logging-clear-handlers",),
     dict(action="store_true",
          help="Clear all other logging handlers.")),

    (("--no-summary",),
     dict(action="store_false", dest="summary",
          help="""Don't display the summary at the end of the run.""")),

    (("--summary",),
     dict(action="store_true", dest="summary",
          help="""Display the summary at the end of the run.""")),

    (("-o", "--outfile"),
     dict(action="append", dest="outfiles", metavar="FILE",
          help="Write to specified file instead of stdout.")),

    ((),  # -- CONFIGFILE only
     dict(action="append", dest="paths",
          help="Specify default feature paths, used when none are provided.")),

    (("-q", "--quiet"),
     dict(action="store_true",
          help="Alias for --no-snippets --no-source.")),

    (("-s", "--no-source"),
     dict(action="store_false", dest="show_source",
          help="""Don't print the file and line of the step definition with the
                  steps.""")),

    (("--show-source",),
     dict(action="store_true", dest="show_source",
          help="""Print the file and line of the step
                  definition with the steps. This is the default
                  behaviour. This switch is used to override a
                  configuration file setting.""")),

    (("--stage",),
     dict(help="""Defines the current test stage.
                  The test stage name is used as name prefix for the environment
                  file and the steps directory (instead of default path names).
                  """)),

    (("--stop",),
     dict(action="store_true",
          help="Stop running tests at the first failure.")),

    # -- DISABLE-UNUSED-OPTION: Not used anywhere.
    # (("-S", "--strict"),
    # dict(action="store_true",
    #    help="Fail if there are any undefined or pending steps.")),

    ((),  # -- CONFIGFILE only
     dict(dest="default_tags", metavar="TAG_EXPRESSION",
          help="""Define default tags when non are provided.
                  See --tags for more information.""")),

    (("-t", "--tags"),
     dict(action="append", metavar="TAG_EXPRESSION",
          help="""Only execute features or scenarios with tags
                  matching TAG_EXPRESSION. Pass "--tags-help" for
                  more information.""",
          config_help="""Only execute certain features or scenarios based
                         on the tag expression given. See below for how to code
                         tag expressions in configuration files.""")),

    (("-T", "--no-timings"),
     dict(action="store_false", dest="show_timings",
          help="""Don't print the time taken for each step.""")),

    (("--show-timings",),
     dict(action="store_true", dest="show_timings",
          help="""Print the time taken, in seconds, of each step after the
                  step has completed. This is the default behaviour. This
                  switch is used to override a configuration file
                  setting.""")),

    (("-v", "--verbose"),
     dict(action="store_true",
          help="Show the files and features loaded.")),

    (("-w", "--wip"),
     dict(action="store_true",
          help="""Only run scenarios tagged with "wip". Additionally: use the
                  "plain" formatter, do not capture stdout or logging output
                  and stop at the first failure.""")),

    (("-x", "--expand"),
     dict(action="store_true",
          help="Expand scenario outline tables in output.")),

    (("--lang",),
     dict(metavar="LANG",
          help="Use keywords for a language other than English.")),

    (("--lang-list",),
     dict(action="store_true",
          help="List the languages available for --lang.")),

    (("--lang-help",),
     dict(metavar="LANG",
          help="List the translations accepted for one language.")),

    (("--tags-help",),
     dict(action="store_true",
          help="Show help for tag expressions.")),

    (("--version",),
     dict(action="store_true", help="Show version.")),
]

# -- OPTIONS: With raw value access semantics in configuration file.
raw_value_options = frozenset([
    "logging_format",
    "logging_datefmt",
    # -- MAYBE: "scenario_outline_annotation_schema",
])


def read_configuration(path):
    # pylint: disable=too-many-locals, too-many-branches
    config = ConfigParser()
    config.optionxform = str    # -- SUPPORT: case-sensitive keys
    config.read(path)
    config_dir = os.path.dirname(path)
    result = {}
    for fixed, keywords in options:
        if "dest" in keywords:
            dest = keywords["dest"]
        else:
            for opt in fixed:
                if opt.startswith("--"):
                    dest = opt[2:].replace("-", "_")
                else:
                    assert len(opt) == 2
                    dest = opt[1:]
        if dest in "tags_help lang_list lang_help version".split():
            continue
        if not config.has_option("behave_bo", dest):
            continue
        action = keywords.get("action", "store")
        if action == "store":
            use_raw_value = dest in raw_value_options
            result[dest] = config.get("behave_bo", dest, raw=use_raw_value)
        elif action in ("store_true", "store_false"):
            result[dest] = config.getboolean("behave_bo", dest)
        elif action == "append":
            if dest == "userdata_defines":
                continue    # -- SKIP-CONFIGFILE: Command-line only option.
            result[dest] = \
                [s.strip() for s in config.get("behave_bo", dest).splitlines()]
        else:
            raise ValueError('action "%s" not implemented' % action)

    # -- STEP: format/outfiles coupling
    if "format" in result:
        # -- OPTIONS: format/outfiles are coupled in configuration file.
        formatters = result["format"]
        formatter_size = len(formatters)
        outfiles = result.get("outfiles", [])
        outfiles_size = len(outfiles)
        if outfiles_size < formatter_size:
            for formatter_name in formatters[outfiles_size:]:
                outfile = "%s.output" % formatter_name
                outfiles.append(outfile)
            result["outfiles"] = outfiles
        elif len(outfiles) > formatter_size:
            print("CONFIG-ERROR: Too many outfiles (%d) provided." %
                  outfiles_size)
            result["outfiles"] = outfiles[:formatter_size]

    for paths_name in ("paths", "outfiles"):
        if paths_name in result:
            # -- Evaluate relative paths relative to location.
            # NOTE: Absolute paths are preserved by os.path.join().
            paths = result[paths_name]
            result[paths_name] = \
                [os.path.normpath(os.path.join(config_dir, p)) for p in paths]

    # -- STEP: Special additional configuration sections.
    # SCHEMA: config_section: data_name
    special_config_section_map = {
        "behave_bo.formatters": "more_formatters",
        "behave_bo.userdata":   "userdata",
    }
    for section_name, data_name in special_config_section_map.items():
        result[data_name] = {}
        if config.has_section(section_name):
            result[data_name].update(config.items(section_name))

    return result


def config_filenames():
    paths = ["./", os.path.expanduser("~")]
    if sys.platform in ("cygwin", "win32") and "APPDATA" in os.environ:
        paths.append(os.path.join(os.environ["APPDATA"]))

    for path in reversed(paths):
        for filename in reversed(
                ("behave_bo.ini", ".behaverc", "setup.cfg", "tox.ini")):
            filename = os.path.join(path, filename)
            if os.path.isfile(filename):
                yield filename


def load_configuration(defaults, verbose=False):
    for filename in config_filenames():
        if verbose:
            print('Loading config defaults from "%s"' % filename)
        defaults.update(read_configuration(filename))

    if verbose:
        print("Using defaults:")
        for k, v in defaults.items():
            print("%15s %s" % (k, v))


def setup_parser():
    # construct the parser
    # usage = "%(prog)s [options] [ [FILE|DIR|URL][:LINE[:LINE]*] ]+"
    usage = "%(prog)s [options] [ [DIR|FILE|FILE:LINE] ]+"
    description = """\
    Run a number of feature tests with behave_bo."""
    more = """
    EXAMPLES:
    behave_bo features/
    behave_bo features/one.feature features/two.feature
    behave_bo features/one.feature:10
    behave_bo @features.txt
    """
    parser = argparse.ArgumentParser(usage=usage, description=description)
    for fixed, keywords in options:
        if not fixed:
            continue    # -- CONFIGFILE only.
        if "config_help" in keywords:
            keywords = dict(keywords)
            del keywords["config_help"]
        parser.add_argument(*fixed, **keywords)
    parser.add_argument("paths", nargs="*",
                        help="Feature directory, file or file location (FILE:LINE).")
    return parser


class Configuration:
    """Configuration object for behave_bo and behave_bo runners."""
    # pylint: disable=too-many-instance-attributes
    defaults = dict(
        color=sys.platform != "win32",
        show_snippets=True,
        show_skipped=True,
        dry_run=False,
        show_source=True,
        show_timings=True,
        stdout_capture=True,
        stderr_capture=True,
        log_capture=True,
        logging_format="%(levelname)s:%(name)s:%(message)s",
        logging_level=logging.INFO,
        steps_catalog=False,
        summary=True,
        junit=False,
        stage=None,
        userdata={},
        # -- SPECIAL:
        default_format="pretty",    # -- Used when no formatters are configured.
        default_tags="",            # -- Used when no tags are defined.
        scenario_outline_annotation_schema="{name} -- @{row.id} {examples.name}"
    )
    cmdline_only_options = set("userdata_defines")

    def __init__(self, command_args=None, load_config=True, verbose=None,
                 **kwargs):
        """
        Constructs a behave_bo configuration object.
          * loads the configuration defaults (if needed).
          * process the command-line args
          * store the configuration results

        :param command_args: Provide command args (as sys.argv).
            If command_args is None, sys.argv[1:] is used.
        :type command_args: list<str>, str
        :param load_config: Indicate if configfile should be loaded (=true)
        :param verbose: Indicate if diagnostic output is enabled
        :param kwargs:  Used to hand-over/overwrite default values.
        """
        # pylint: disable=too-many-branches, too-many-statements
        if command_args is None:
            command_args = sys.argv[1:]
        elif isinstance(command_args, str):
            encoding = select_best_encoding() or "utf-8"
            if isinstance(command_args, bytes):
                command_args = command_args.decode(encoding)
            command_args = shlex.split(command_args)
        elif isinstance(command_args, (list, tuple)):
            command_args = to_texts(command_args)

        if verbose is None:
            # -- AUTO-DISCOVER: Verbose mode from command-line args.
            verbose = ("-v" in command_args) or ("--verbose" in command_args)

        self.version = None
        self.tags_help = None
        self.lang_list = None
        self.lang_help = None
        self.default_tags = None
        self.junit = None
        self.logging_format = None
        self.logging_datefmt = None
        self.name = None
        self.scope = None
        self.steps_catalog = None
        self.userdata = None
        self.wip = None

        defaults = self.defaults.copy()
        for name, value in kwargs.items():
            defaults[name] = value
        self.defaults = defaults
        self.formatters = []
        self.reporters = []
        self.name_re = None
        self.outputs = []
        self.include_re = None
        self.exclude_re = None
        self.scenario_outline_annotation_schema = None  # pylint: disable=invalid-name
        self.steps_dir = "steps"
        self.environment_file = "environment.py"
        self.userdata_defines = None
        self.more_formatters = None
        if load_config:
            load_configuration(self.defaults, verbose=verbose)
        parser = setup_parser()
        parser.set_defaults(**self.defaults)
        args = parser.parse_args(command_args)
        for key, value in args.__dict__.items():
            if key.startswith("_") and key not in self.cmdline_only_options:
                continue
            setattr(self, key, value)

        self.paths = [os.path.normpath(path) for path in self.paths]
        self.setup_outputs(args.outfiles)

        if self.steps_catalog:
            # -- SHOW STEP-CATALOG: As step summary.
            self.default_format = "steps.catalog"
            self.format = ["steps.catalog"]
            self.dry_run = True
            self.summary = False
            self.show_skipped = False
            self.quiet = True

        if self.wip:
            # Only run scenarios tagged with "wip".
            # Additionally:
            #  * use the "plain" formatter (per default)
            #  * do not capture stdout or logging output and
            #  * stop at the first failure.
            self.default_format = "plain"
            self.tags = ["wip"] + self.default_tags.split()
            self.color = False
            self.stop = True
            self.log_capture = False
            self.stdout_capture = False

        self.tags = TagExpression(self.tags or self.default_tags.split())

        if self.quiet:
            self.show_source = False
            self.show_snippets = False

        if self.exclude_re:
            self.exclude_re = re.compile(self.exclude_re)

        if self.include_re:
            self.include_re = re.compile(self.include_re)
        if self.name:
            # -- SELECT: Scenario-by-name, build regular expression.
            self.name_re = self.build_name_re(self.name)

        if self.stage is None:  # pylint: disable=access-member-before-definition
            # -- USE ENVIRONMENT-VARIABLE, if stage is undefined.
            self.stage = os.environ.get("BEHAVE_STAGE", None)
        self.setup_stage(self.stage)
        self.setup_model()
        self.setup_userdata()

        # -- FINALLY: Setup Reporters and Formatters
        # NOTE: Reporters and Formatters can now use userdata information.
        if self.junit:
            # Buffer the output (it will be put into Junit report)
            self.stdout_capture = True
            self.stderr_capture = True
            self.log_capture = True
            self.reporters.append(JUnitReporter(self))
        if self.summary:
            self.reporters.append(SummaryReporter(self))

        self.setup_formats()
        unknown_formats = self.collect_unknown_formats()
        if unknown_formats:
            parser.error("format=%s is unknown" % ", ".join(unknown_formats))


    def setup_outputs(self, args_outfiles=None):
        if self.outputs:
            assert not args_outfiles, "ONLY-ONCE"
            return

        # -- NORMAL CASE: Setup only initially (once).
        if not args_outfiles:
            self.outputs.append(StreamOpener(stream=sys.stdout))
        else:
            for outfile in args_outfiles:
                if outfile and outfile != "-":
                    self.outputs.append(StreamOpener(outfile))
                else:
                    self.outputs.append(StreamOpener(stream=sys.stdout))

    def setup_formats(self):
        """Register more, user-defined formatters by name."""
        if self.more_formatters:
            for name, scoped_class_name in self.more_formatters.items():
                _format_registry.register_as(name, scoped_class_name)

    def collect_unknown_formats(self):
        unknown_formats = []
        if self.format:
            for format_name in self.format:
                if (format_name == "help" or
                        _format_registry.is_formatter_valid(format_name)):
                    continue
                unknown_formats.append(format_name)
        return unknown_formats

    @staticmethod
    def build_name_re(names):
        """
        Build regular expression for scenario selection by name
        by using a list of name parts or name regular expressions.

        :param names: List of name parts or regular expressions (as text).
        :return: Compiled regular expression to use.
        """
        # -- NOTE: re.LOCALE is removed in Python 3.6 (deprecated in Python 3.5)
        # flags = (re.UNICODE | re.LOCALE)
        # -- ENSURE: Names are all unicode/text values (for issue #606).
        names = to_texts(names)
        pattern = "|".join(names)
        return re.compile(pattern, flags=re.UNICODE)

    def exclude(self, filename):
        if isinstance(filename, FileLocation):
            filename = str(filename)

        if self.include_re and self.include_re.search(filename) is None:
            return True
        if self.exclude_re and self.exclude_re.search(filename) is not None:
            return True
        return False

    def setup_logging(self, level=None, configfile=None, **kwargs):
        """
        Support simple setup of logging subsystem.
        Ensures that the logging level is set.
        But note that the logging setup can only occur once.

        SETUP MODES:
          * :func:`logging.config.fileConfig()`, if ``configfile`` is provided.
          * :func:`logging.basicConfig()`, otherwise.

        .. code-block: python
            # -- FILE: features/environment.py
            def before_all(context):
                context.config.setup_logging()

        :param level:       Logging level of root logger.
                            If None, use :attr:`logging_level` value.
        :param configfile:  Configuration filename for fileConfig() setup.
        :param kwargs:      Passed to :func:`logging.basicConfig()`
        """
        if level is None:
            level = self.logging_level      # pylint: disable=no-member

        if configfile:
            from logging.config import (
                fileConfig,
            )
            fileConfig(configfile)
        else:
            # pylint: disable=no-member
            format_ = kwargs.pop("format", self.logging_format)
            datefmt = kwargs.pop("datefmt", self.logging_datefmt)
            logging.basicConfig(format=format_, datefmt=datefmt, **kwargs)
        # -- ENSURE: Default log level is set
        #    (even if logging subsystem is already configured).
        logging.getLogger().setLevel(level)

    def setup_model(self):
        if self.scenario_outline_annotation_schema:
            name_schema = str(self.scenario_outline_annotation_schema)
            ScenarioOutline.annotation_schema = name_schema.strip()

    def setup_stage(self, stage=None):
        """Setup the test stage that selects a different set of
        steps and environment implementations.

        :param stage:   Name of current test stage (as string or None).

        EXAMPLE::

            # -- SETUP DEFAULT TEST STAGE (unnamed):
            config = Configuration()
            config.setup_stage()
            assert config.steps_dir == "steps"
            assert config.environment_file == "environment.py"

            # -- SETUP PRODUCT TEST STAGE:
            config.setup_stage("product")
            assert config.steps_dir == "product_steps"
            assert config.environment_file == "product_environment.py"
        """
        steps_dir = "steps"
        environment_file = "environment.py"
        if stage:
            # -- USE A TEST STAGE: Select different set of implementations.
            prefix = stage + "_"
            steps_dir = prefix + steps_dir
            environment_file = prefix + environment_file
        self.steps_dir = steps_dir
        self.environment_file = environment_file

    def setup_userdata(self):
        if not isinstance(self.userdata, UserData):
            self.userdata = UserData(self.userdata)
        if self.userdata_defines:
            # -- ENSURE: Cmd-line overrides configuration file parameters.
            self.userdata.update(self.userdata_defines)

    def update_userdata(self, data):
        """Update userdata with data and reapply userdata defines (cmdline).
        :param data:  Provides (partial) userdata (as dict)
        """
        self.userdata.update(data)
        if self.userdata_defines:
            # -- REAPPLY: Cmd-line defines (override configuration file data).
            self.userdata.update(self.userdata_defines)



from behave_bo.formatter.base import (
    Formatter,
)


class NullFormatter(Formatter):
    """
    Provides formatter that does not output anything.
    Implements the NULL pattern for a formatter (similar like: /dev/null).
    """
    name = "null"
    description = "Provides formatter that does not output anything."



"""
Collects data how often a tag count is used and where.

EXAMPLE:

    $ behave_bo --dry-run -f tag_counts features/
"""

from behave_bo.formatter.base import (
    Formatter,
)
from behave_bo.textutil import (
    compute_words_maxsize,
    text as _text,
)


# -----------------------------------------------------------------------------
# CLASS: AbstractTagsFormatter
# -----------------------------------------------------------------------------
class AbstractTagsFormatter(Formatter):
    """
    Abstract base class for formatter that collect information on tags.

    .. note::
        Supports dry-run mode for faster feedback.
    """
    with_tag_inheritance = False

    def __init__(self, stream_opener, config):
        super().__init__(stream_opener, config)
        self.tag_counts = {}
        self._uri = None
        self._feature_tags = None
        self._scenario_outline_tags = None

    # -- Formatter API:
    def uri(self, uri):
        self._uri = uri

    def feature(self, feature):
        self._feature_tags = feature.tags
        self.record_tags(feature.tags, feature)

    def scenario(self, scenario):
        tags = set(scenario.tags)
        if self.with_tag_inheritance:
            tags.update(self._feature_tags)
        self.record_tags(tags, scenario)

    def close(self):
        """Emit tag count reports."""
        # -- ENSURE: Output stream is open.
        self.stream = self.open()
        self.report_tags()
        self.close_stream()

    # -- SPECIFIC API:
    def record_tags(self, tags, model_element):
        for tag in tags:
            if tag not in self.tag_counts:
                self.tag_counts[tag] = []
            self.tag_counts[tag].append(model_element)

    def report_tags(self):
        raise NotImplementedError


# -----------------------------------------------------------------------------
# CLASS: TagsFormatter
# -----------------------------------------------------------------------------
class TagsFormatter(AbstractTagsFormatter):
    """
    Formatter that collects information:

      * which tags exist
      * how often a tag is used (counts)
      * usage context/category: feature, scenario, ...

    .. note::
        Supports dry-run mode for faster feedback.
    """
    name = "tags"
    description = "Shows tags (and how often they are used)."
    with_tag_inheritance = False
    show_ordered_by_usage = False

    def report_tags(self):
        self.report_tag_counts()
        if self.show_ordered_by_usage:
            self.report_tag_counts_by_usage()

    @staticmethod
    def get_tag_count_details(tag_count):
        details = {}
        for element in tag_count:
            category = element.keyword.lower()
            if category not in details:
                details[category] = 0
            details[category] += 1

        parts = []
        if len(details) == 1:
            parts.append(list(details.keys())[0])
        else:
            for category in sorted(details):
                text = "%s: %d" % (category, details[category])
                parts.append(text)
        return ", ".join(parts)

    def report_tag_counts(self):
        # -- PREPARE REPORT:
        ordered_tags = sorted(list(self.tag_counts.keys()))
        tag_maxsize = compute_words_maxsize(ordered_tags)
        schema = "  @%-" + _text(tag_maxsize) + "s %4d    (used for %s)\n"

        # -- EMIT REPORT:
        self.stream.write("TAG COUNTS (alphabetically sorted):\n")
        for tag in ordered_tags:
            tag_data = self.tag_counts[tag]
            counts = len(tag_data)
            details = self.get_tag_count_details(tag_data)
            self.stream.write(schema % (tag, counts, details))
        self.stream.write("\n")

    def report_tag_counts_by_usage(self):
        # -- PREPARE REPORT:
        compare_tag_counts_size = lambda x: len(self.tag_counts[x])
        ordered_tags = sorted(list(self.tag_counts.keys()),
                              key=compare_tag_counts_size)
        tag_maxsize = compute_words_maxsize(ordered_tags)
        schema = "  @%-" + _text(tag_maxsize) + "s %4d    (used for %s)\n"

        # -- EMIT REPORT:
        self.stream.write("TAG COUNTS (most often used first):\n")
        for tag in ordered_tags:
            tag_data = self.tag_counts[tag]
            counts = len(tag_data)
            details = self.get_tag_count_details(tag_data)
            self.stream.write(schema % (tag, counts, details))
        self.stream.write("\n")


# -----------------------------------------------------------------------------
# CLASS: TagsLocationFormatter
# -----------------------------------------------------------------------------
class TagsLocationFormatter(AbstractTagsFormatter):
    """
    Formatter that collects information:

      * which tags exist
      * where the tags are used (location)

    .. note::
        Supports dry-run mode for faster feedback.
    """
    name = "tags.location"
    description = "Shows tags and the location where they are used."
    with_tag_inheritance = False

    def report_tags(self):
        self.report_tags_by_locations()

    def report_tags_by_locations(self):
        # -- PREPARE REPORT:
        locations = set()
        for tag_elements in self.tag_counts.values():
            locations.update([str(x.location) for x in tag_elements])
        location_column_size = compute_words_maxsize(locations)
        schema = "    %-" + _text(location_column_size) + "s   %s\n"

        # -- EMIT REPORT:
        self.stream.write("TAG LOCATIONS (alphabetically ordered):\n")
        for tag in sorted(self.tag_counts):
            self.stream.write("  @%s:\n" % tag)
            for element in self.tag_counts[tag]:
                info = f"{element.keyword}: {element.name}"
                self.stream.write(schema % (element.location, info))
            self.stream.write("\n")
        self.stream.write("\n")



"""
This module provides Runner class to run behave_bo feature files (or model elements).
"""

import contextlib
import datetime
import itertools
import json
import multiprocessing
import os
import os.path
import random
import sys
import traceback
import warnings
import weakref
from functools import (
    cached_property,
)
from glob import (
    glob,
)

from behave_bo import (
    register_type,
)
from behave_bo._types import (
    ExceptionUtil,
)
from behave_bo.capture import (
    CaptureController,
)
from behave_bo.configuration import (
    ConfigError,
)
from behave_bo.consts import (
    scenario_group_prefix,
)
from behave_bo.formatter._registry import (
    make_formatters,
)
from behave_bo.loggers import (
    tests_logger,
)
from behave_bo.matchers import (
    CaseSensitiveParser,
    ParseMatcher,
)
from behave_bo.reporter.junit import (
    JunitReportsCombiner,
)
from behave_bo.runner_util import (
    PathManager,
    collect_feature_locations,
    exec_file,
    load_step_modules,
    parse_features,
)
from behave_bo.step_registry import (
    registry as the_step_registry,
)


class CleanupError(RuntimeError):
    pass


class ContextMaskWarning(UserWarning):
    """Raised if a context variable is being overwritten in some situations.

    If the variable was originally set by user code then this will be raised if
    *behave_bo* overwites the value.

    If the variable was originally set by *behave_bo* then this will be raised if
    user code overwites the value.
    """
    pass


class Context:
    """Hold contextual information during the running of tests.

    This object is a place to store information related to the tests you're
    running. You may add arbitrary attributes to it of whatever value you need.

    During the running of your tests the object will have additional layers of
    namespace added and removed automatically. There is a "root" namespace and
    additional namespaces for features and scenarios.

    Certain names are used by *behave_bo*; be wary of using them yourself as
    *behave_bo* may overwrite the value you set. These names are:

    .. attribute:: feature

      This is set when we start testing a new feature and holds a
      :class:`~behave_bo.model.Feature`. It will not be present outside of a
      feature (i.e. within the scope of the environment before_all and
      after_all).

    .. attribute:: scenario

      This is set when we start testing a new scenario (including the
      individual scenarios of a scenario outline) and holds a
      :class:`~behave_bo.model.Scenario`. It will not be present outside of the
      scope of a scenario.

    .. attribute:: tags

      The current set of active tags (as a Python set containing instances of
      :class:`~behave_bo.model.Tag` which are basically just glorified strings)
      combined from the feature and scenario. This attribute will not be
      present outside of a feature scope.

    .. attribute:: aborted

      This is set to true in the root namespace when the user aborts a test run
      (:exc:`KeyboardInterrupt` exception). Initially: False.

    .. attribute:: failed

      This is set to true in the root namespace as soon as a step fails.
      Initially: False.

    .. attribute:: table

      This is set at the step level and holds any :class:`~behave_bo.model.Table`
      associated with the step.

    .. attribute:: text

      This is set at the step level and holds any multiline text associated
      with the step.

    .. attribute:: config

      The configuration of *behave_bo* as determined by configuration files and
      command-line options. The attributes of this object are the same as the
      `configuration file section names`_.

    .. attribute:: active_outline

      This is set for each scenario in a scenario outline and references the
      :class:`~behave_bo.model.Row` that is active for the current scenario. It is
      present mostly for debugging, but may be useful otherwise.

    .. attribute:: log_capture

      If logging capture is enabled then this attribute contains the captured
      logging as an instance of :class:`~behave_bo.log_capture.LoggingCapture`.
      It is not present if logging is not being captured.

    .. attribute:: stdout_capture

      If stdout capture is enabled then this attribute contains the captured
      output as a StringIO instance. It is not present if stdout is not being
      captured.

    .. attribute:: stderr_capture

      If stderr capture is enabled then this attribute contains the captured
      output as a StringIO instance. It is not present if stderr is not being
      captured.

    If an attempt made by user code to overwrite one of these variables, or
    indeed by *behave_bo* to overwite a user-set variable, then a
    :class:`behave_bo.runner.ContextMaskWarning` warning will be raised.

    You may use the "in" operator to test whether a certain value has been set
    on the context, for example:

        "feature" in context

    checks whether there is a "feature" value in the context.

    Values may be deleted from the context using "del" but only at the level
    they are set. You can't delete a value set by a feature at a scenario level
    but you can delete a value set for a scenario in that scenario.

    .. _`configuration file section names`: behave_bo.html#configuration-files
    """
    # pylint: disable=too-many-instance-attributes
    BEHAVE = "behave_bo"
    USER = "user"
    FAIL_ON_CLEANUP_ERRORS = True

    def __init__(self, runner):
        self._runner = weakref.proxy(runner)
        self._config = runner.config
        d = self._root = {
            "aborted": False,
            "failed": False,
            "config": self._config,
            "active_outline": None,
            "cleanup_errors": 0,
            "@cleanups": [],  # -- REQUIRED-BY: before_all() hook
            "@layer": "testrun",
        }
        self._stack = [d]
        self._record = {}
        self._origin = {}
        self._mode = self.BEHAVE
        self.feature = None
        # -- RECHECK: If needed
        self.text = None
        self.table = None
        self.stdout_capture = None
        self.stderr_capture = None
        self.log_capture = None
        self.fail_on_cleanup_errors = self.FAIL_ON_CLEANUP_ERRORS

    @staticmethod
    def ignore_cleanup_error(context, cleanup_func, exception):
        pass

    @staticmethod
    def print_cleanup_error(context, cleanup_func, exception):
        cleanup_func_name = getattr(cleanup_func, "__name__", None)
        if not cleanup_func_name:
            cleanup_func_name = "%r" % cleanup_func
        print("CLEANUP-ERROR in %s: %s: %s" %
              (cleanup_func_name, exception.__class__.__name__, exception))
        traceback.print_exc(file=sys.stdout)
        # MAYBE: context._dump(pretty=True, prefix="Context: ")
        # -- MARK: testrun as FAILED
        # context._set_root_attribute("failed", True)

    def _do_cleanups(self):
        """Execute optional cleanup functions when stack frame is popped.
        A user can add a user-specified handler for cleanup errors.

        .. code-block:: python

            # -- FILE: features/environment.py
            def cleanup_database(database):
                pass

            def handle_cleanup_error(context, cleanup_func, exception):
                pass

            def before_all(context):
                context.on_cleanup_error = handle_cleanup_error
                context.add_cleanup(cleanup_database, the_database)
        """
        # -- BEST-EFFORT ALGORITHM: Tries to perform all cleanups.
        assert self._stack, "REQUIRE: Non-empty stack"
        current_layer = self._stack[0]
        cleanup_funcs = current_layer.get("@cleanups", [])
        on_cleanup_error = getattr(self, "on_cleanup_error",
                                   self.print_cleanup_error)
        context = self
        cleanup_errors = []
        for cleanup_func in reversed(cleanup_funcs):
            try:
                cleanup_func()
            except Exception as e:  # pylint: disable=broad-except
                # pylint: disable=protected-access
                context._root["cleanup_errors"] += 1
                cleanup_errors.append(sys.exc_info())
                on_cleanup_error(context, cleanup_func, e)

        if self.fail_on_cleanup_errors and cleanup_errors:
            first_cleanup_erro_info = cleanup_errors[0]
            del cleanup_errors  # -- ENSURE: Release other exception frames.

    def _push(self, layer_name=None):
        """Push a new layer on the context stack.
        HINT: Use layer_name values: "scenario", "feature", "testrun".

        :param layer_name:   Layer name to use (or None).
        """
        initial_data = {"@cleanups": []}
        if layer_name:
            initial_data["@layer"] = layer_name
        self._stack.insert(0, initial_data)

    def _pop(self):
        """Pop the current layer from the context stack.
        Performs any pending cleanups, registered for this layer.
        """
        try:
            self._do_cleanups()
        finally:
            # -- ENSURE: Layer is removed even if cleanup-errors occur.
            self._stack.pop(0)

    def _use_with_behave_mode(self):
        """Provides a context manager for using the context in BEHAVE mode."""
        return use_context_with_mode(self, Context.BEHAVE)

    def use_with_user_mode(self):
        """Provides a context manager for using the context in USER mode."""
        return use_context_with_mode(self, Context.USER)

    def _set_root_attribute(self, attr, value):
        for frame in self.__dict__["_stack"]:
            if frame is self.__dict__["_root"]:
                continue
            if attr in frame:
                record = self.__dict__["_record"][attr]
                params = {
                    "attr": attr,
                    "filename": record[0],
                    "line": record[1],
                    "function": record[3],
                }
                self._emit_warning(attr, params)

        self.__dict__["_root"][attr] = value
        if attr not in self._origin:
            self._origin[attr] = self._mode

    def _emit_warning(self, attr, params):
        msg = ""
        if self._mode is self.BEHAVE and self._origin[attr] is not self.BEHAVE:
            msg = "behave_bo runner is masking context attribute '%(attr)s' " \
                  "originally set in %(function)s (%(filename)s:%(line)s)"
        elif self._mode is self.USER:
            if self._origin[attr] is not self.USER:
                msg = "user code is masking context attribute '%(attr)s' " \
                      "originally set by behave_bo"
            elif self._config.verbose:
                msg = "user code is masking context attribute " \
                      "'%(attr)s'; see the tutorial for what this means"
        if msg:
            msg = msg % params
            warnings.warn(msg, ContextMaskWarning, stacklevel=3)

    def _dump(self, pretty=False, prefix="  "):
        for level, frame in enumerate(self._stack):
            print("%sLevel %d" % (prefix, level))
            if pretty:
                for name in sorted(frame.keys()):
                    value = frame[name]
                    print("%s  %-15s = %r" % (prefix, name, value))
            else:
                print(prefix + repr(frame))

    def __getattr__(self, attr):
        if attr[0] == "_":
            return self.__dict__[attr]
        for frame in self._stack:
            if attr in frame:
                return frame[attr]
        msg = "'{0}' object has no attribute '{1}'"
        msg = msg.format(self.__class__.__name__, attr)
        raise AttributeError(msg)

    def __setattr__(self, attr, value):
        if attr[0] == "_":
            self.__dict__[attr] = value
            return

        for frame in self._stack[1:]:
            if attr in frame:
                record = self._record[attr]
                params = {
                    "attr": attr,
                    "filename": record[0],
                    "line": record[1],
                    "function": record[3],
                }
                self._emit_warning(attr, params)

        stack_limit = 2
        stack_frame = traceback.extract_stack(limit=stack_limit)[0]
        self._record[attr] = stack_frame
        frame = self._stack[0]
        frame[attr] = value
        if attr not in self._origin:
            self._origin[attr] = self._mode

    def __delattr__(self, attr):
        frame = self._stack[0]
        if attr in frame:
            del frame[attr]
            del self._record[attr]
        else:
            msg = "'{0}' object has no attribute '{1}' at the current level"
            msg = msg.format(self.__class__.__name__, attr)
            raise AttributeError(msg)

    def __contains__(self, attr):
        if attr[0] == "_":
            return attr in self.__dict__
        for frame in self._stack:
            if attr in frame:
                return True
        return False

    def execute_steps(self, steps_text):
        """The steps identified in the "steps" text string will be parsed and
        executed in turn just as though they were defined in a feature file.

        If the execute_steps call fails (either through error or failure
        assertion) then the step invoking it will need to catch the resulting
        exceptions.

        :param steps_text:  Text with the Gherkin steps to execute (as string).
        :returns: True, if the steps executed successfully.
        :raises: AssertionError, if a step failure occurs.
        :raises: ValueError, if invoked without a feature context.
        """
        assert isinstance(steps_text, str), "Steps must be unicode."
        if not self.feature:
            raise ValueError("execute_steps() called outside of feature")

        # -- PREPARE: Save original context data for current step.
        # Needed if step definition that called this method uses .table/.text
        original_table = getattr(self, "table", None)
        original_text = getattr(self, "text", None)

        self.feature.parser.variant = "steps"
        steps = self.feature.parser.parse_steps(steps_text)
        with self._use_with_behave_mode():
            for step in steps:
                passed = step.run(self._runner, quiet=True, capture=False)
                if not passed:
                    # -- ISSUE #96: Provide more substep info to diagnose problem.
                    step_line = f"{step.keyword} {step.name}"
                    message = "%s SUB-STEP: %s" % \
                              (step.status.name.upper(), step_line)
                    if step.error_message:
                        message += "\nSubstep info: %s\n" % step.error_message
                        message += "Traceback (of failed substep):\n"
                        message += "".join(traceback.format_tb(step.exc_traceback))
                    # message += u"\nTraceback (of context.execute_steps()):"
                    assert False, message

            # -- FINALLY: Restore original context data for current step.
            self.table = original_table
            self.text = original_text
        return True

    def add_cleanup(self, cleanup_func, *args, **kwargs):
        """Adds a cleanup function that is called when :meth:`Context._pop()`
        is called. This is intended for user-cleanups.

        :param cleanup_func:    Callable function
        :param args:            Args for cleanup_func() call (optional).
        :param kwargs:          Kwargs for cleanup_func() call (optional).
        """
        # MAYBE:
        assert callable(cleanup_func), "REQUIRES: callable(cleanup_func)"
        assert self._stack
        if args or kwargs:
            def internal_cleanup_func():
                cleanup_func(*args, **kwargs)
        else:
            internal_cleanup_func = cleanup_func

        current_frame = self._stack[0]
        if cleanup_func not in current_frame["@cleanups"]:
            # -- AVOID DUPLICATES:
            current_frame["@cleanups"].append(internal_cleanup_func)


@contextlib.contextmanager
def use_context_with_mode(context, mode):
    """Switch context to BEHAVE or USER mode.
    Provides a context manager for switching between the two context modes.

    .. sourcecode:: python

        context = Context()
        with use_context_with_mode(context, Context.BEHAVE):
            ...     # Do something
        # -- POSTCONDITION: Original context._mode is restored.

    :param context:  Context object to use.
    :param mode:     Mode to apply to context object.
    """
    # pylint: disable=protected-access
    assert mode in (Context.BEHAVE, Context.USER)
    current_mode = context._mode
    try:
        context._mode = mode
        yield
    finally:
        # -- RESTORE: Initial current_mode
        #    Even if an AssertionError/Exception is raised.
        context._mode = current_mode


@contextlib.contextmanager
def scoped_context_layer(context, layer_name=None):
    """Provides context manager for context layer (push/do-something/pop cycle).

    .. code-block::

        with scoped_context_layer(context):
            the_fixture = use_fixture(foo, context, name="foo_42")
    """
    # pylint: disable=protected-access
    try:
        context._push(layer_name)
        yield context
    finally:
        context._pop()


def path_getrootdir(path):
    """
    Extract rootdir from path in a platform independent way.

    POSIX-PATH EXAMPLE:
        rootdir = path_getrootdir("/foo/bar/one.feature")
        assert rootdir == "/"

    WINDOWS-PATH EXAMPLE:
        rootdir = path_getrootdir("D:\\foo\\bar\\one.feature")
        assert rootdir == r"D:\"
    """
    drive, _ = os.path.splitdrive(path)
    if drive:
        # -- WINDOWS:
        return drive + os.path.sep
    # -- POSIX:
    return os.path.sep


class ModelRunner:
    """
    Test runner for a behave_bo model (features).
    Provides the core functionality of a test runner and
    the functional API needed by model elements.

    .. attribute:: aborted

          This is set to true when the user aborts a test run
          (:exc:`KeyboardInterrupt` exception). Initially: False.
          Stored as derived attribute in :attr:`Context.aborted`.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(self, config, features=None, step_registry=None):
        self.config = config
        self.features = features or []
        self.hooks = {}
        self.formatters = []
        self.undefined_steps = []
        self.step_registry = step_registry
        self.capture_controller = CaptureController(config)

        self.context = None
        self.feature = None
        self.hook_failures = 0

    # @property
    def _get_aborted(self):
        value = False
        if self.context:
            value = self.context.aborted
        return value

    # @aborted.setter
    def _set_aborted(self, value):
        # pylint: disable=protected-access
        assert self.context, "REQUIRE: context, but context=%r" % self.context
        self.context._set_root_attribute("aborted", bool(value))

    aborted = property(_get_aborted, _set_aborted,
                       doc="Indicates that test run is aborted by the user.")

    def run_hook(self, name, context, *args):
        if not self.config.dry_run and (name in self.hooks):
            try:
                with context.use_with_user_mode():
                    self.hooks[name](context, *args)
            # except KeyboardInterrupt:
            #     self.aborted = True
            #     if name not in ("before_all", "after_all"):
            #         raise
            except Exception as e:  # pylint: disable=broad-except
                # -- HANDLE HOOK ERRORS:
                use_traceback = False
                if self.config.verbose:
                    use_traceback = True
                    ExceptionUtil.set_traceback(e)
                extra = ""
                if "tag" in name:
                    extra = "(tag=%s)" % args[0]

                error_text = ExceptionUtil.describe(e, use_traceback).rstrip()
                error_message = f"HOOK-ERROR in {name}{extra}: {error_text}"
                print(error_message)
                self.hook_failures += 1
                if "tag" in name:
                    # -- SCENARIO or FEATURE
                    statement = getattr(context, "scenario", context.feature)
                elif "all" in name:
                    # -- ABORT EXECUTION: For before_all/after_all
                    self.aborted = True
                    statement = None
                else:
                    # -- CASE: feature, scenario, step
                    statement = args[0]

                if statement:
                    # -- CASE: feature, scenario, step
                    statement.hook_failed = True
                    if statement.error_message:
                        # -- NOTE: One exception/failure is already stored.
                        #    Append only error message.
                        statement.error_message += "\n" + error_message
                    else:
                        # -- FIRST EXCEPTION/FAILURE:
                        statement.store_exception_context(e)
                        statement.error_message = error_message

    def setup_capture(self):
        if not self.context:
            self.context = Context(self)
        self.capture_controller.setup_capture(self.context)

    def start_capture(self):
        self.capture_controller.start_capture()

    def stop_capture(self):
        self.capture_controller.stop_capture()

    def teardown_capture(self):
        self.capture_controller.teardown_capture()

    def should_run_feature(self, feature):
        """
        Проверка нужно ли запускать feature.
        """
        return True

    def pre_run_feature(self, feature):
        """Действия перед запуском feature"""
        pass

    def run_model(self, features=None):
        # pylint: disable=too-many-branches
        if not self.context:
            self.context = Context(self)
        if self.step_registry is None:
            self.step_registry = the_step_registry
        if features is None:
            features = self.features

        # -- ENSURE: context.execute_steps() works in weird cases (hooks, ...)
        context = self.context
        self.hook_failures = 0
        self.setup_capture()

        if not self.config.save_db or self.config.parallel:
            # Настроим отдельную БД (которая после завершения удаляется) и подключимся к ней.
            self.config.django_test_runner.setup_databases(
                db_postfix=str(os.getpid()),
            )

        self.run_hook("before_all", context)

        run_feature = not self.aborted
        failed_count = 0
        undefined_steps_initial_size = len(self.undefined_steps)
        for feature in features:
            if not self.should_run_feature(feature):
                # Если не нужно запускать feature, просто пропустим, чтобы не инициировать лишние вызовы reporters.
                continue

            if run_feature:
                try:
                    self.feature = feature
                    for formatter in self.formatters:
                        formatter.uri(feature.filename)

                    self.pre_run_feature(feature)
                    failed = feature.run(self)
                    if failed:
                        failed_count += 1
                        if self.config.stop or self.aborted:
                            # -- FAIL-EARLY: After first failure.
                            run_feature = False
                except KeyboardInterrupt:
                    self.aborted = True
                    failed_count += 1
                    run_feature = False

            # -- ALWAYS: Report run/not-run feature to reporters.
            # REQUIRED-FOR: Summary to keep track of untested features.
            for reporter in self.config.reporters:
                reporter.feature(feature)

        # -- AFTER-ALL:
        # pylint: disable=protected-access, broad-except
        cleanups_failed = False
        self.run_hook("after_all", self.context)
        try:
            self.context._do_cleanups()  # Without dropping the last context layer.
        except Exception:
            cleanups_failed = True

        if self.aborted:
            print("\nABORTED: By user.")
        for formatter in self.formatters:
            formatter.close()
        for reporter in self.config.reporters:
            reporter.end()

        failed = ((failed_count > 0) or self.aborted or (self.hook_failures > 0)
                  or (len(self.undefined_steps) > undefined_steps_initial_size)
                  or cleanups_failed)
        # XXX-MAYBE: or context.failed)
        return failed

    def run(self):
        """
        Implements the run method by running the model.
        """
        self.context = Context(self)
        return self.run_model()


class Runner(ModelRunner):
    """
    Standard test runner for behave_bo:

      * setup paths
      * loads environment hooks
      * loads step definitions
      * select feature files, parses them and creates model (elements)
    """

    def __init__(self, config, **kwargs):
        super().__init__(config)
        self.path_manager = PathManager()
        self.base_dir = None

    def setup_paths(self):
        # pylint: disable=too-many-branches, too-many-statements
        if self.config.paths:
            if self.config.verbose:
                print("Supplied path:",
                      ", ".join('"%s"' % path for path in self.config.paths))
            first_path = self.config.paths[0]
            if hasattr(first_path, "filename"):
                # -- BETTER: isinstance(first_path, FileLocation):
                first_path = first_path.filename
            base_dir = first_path
            if base_dir.startswith("@"):
                # -- USE: behave_bo @features.txt
                base_dir = base_dir[1:]
                file_locations = self.feature_locations()
                if file_locations:
                    base_dir = os.path.dirname(file_locations[0].filename)
            base_dir = os.path.abspath(base_dir)

            # supplied path might be to a feature file
            if os.path.isfile(base_dir):
                if self.config.verbose:
                    print("Primary path is to a file so using its directory")
                base_dir = os.path.dirname(base_dir)
        else:
            if self.config.verbose:
                print('Using default path "./features"')
            base_dir = os.path.abspath("features")

        # Get the root. This is not guaranteed to be "/" because Windows.
        root_dir = path_getrootdir(base_dir)
        new_base_dir = base_dir
        steps_dir = self.config.steps_dir
        environment_file = self.config.environment_file

        while True:
            if self.config.verbose:
                print("Trying base directory:", new_base_dir)

            if os.path.isdir(os.path.join(new_base_dir, steps_dir)):
                break
            if os.path.isfile(os.path.join(new_base_dir, environment_file)):
                break
            if new_base_dir == root_dir:
                break

            new_base_dir = os.path.dirname(new_base_dir)

        if new_base_dir == root_dir:
            if self.config.verbose:
                if not self.config.paths:
                    print('ERROR: Could not find "%s" directory. '
                          'Please specify where to find your features.' % steps_dir)
                else:
                    print('ERROR: Could not find "%s" directory in your '
                          'specified path "%s"' % (steps_dir, base_dir))

            message = f'No {steps_dir} directory in {base_dir!r}'
            raise ConfigError(message)

        base_dir = new_base_dir
        self.config.base_dir = base_dir

        for dirpath, dirnames, filenames in os.walk(base_dir):
            if [fn for fn in filenames if fn.endswith(".feature")]:
                break
        else:
            if self.config.verbose:
                if not self.config.paths:
                    print('ERROR: Could not find any "<name>.feature" files. '
                          'Please specify where to find your features.')
                else:
                    print('ERROR: Could not find any "<name>.feature" files '
                          'in your specified path "%s"' % base_dir)
            raise ConfigError('No feature files in %r' % base_dir)

        self.base_dir = base_dir
        self.path_manager.add(base_dir)
        if not self.config.paths:
            self.config.paths = [base_dir]

        if base_dir != os.getcwd():
            self.path_manager.add(os.getcwd())

    def before_all_default_hook(self, context):
        """
        Default implementation for :func:`before_all()` hook.
        Setup the logging subsystem based on the configuration data.
        """
        # pylint: disable=no-self-use
        context.config.setup_logging()

    def load_hooks(self, filename=None):
        filename = filename or self.config.environment_file
        hooks_path = os.path.join(self.base_dir, filename)
        if os.path.exists(hooks_path):
            exec_file(hooks_path, self.hooks)

        if "before_all" not in self.hooks:
            self.hooks["before_all"] = self.before_all_default_hook

    def load_step_definitions(self, extra_step_paths=None):
        if extra_step_paths is None:
            extra_step_paths = []
        # -- Allow steps to import other stuff from the steps dir
        # NOTE: Default matcher can be overridden in "environment.py" hook.
        steps_dir = os.path.join(self.base_dir, self.config.steps_dir)
        step_paths = [steps_dir] + list(extra_step_paths)
        load_step_modules(step_paths)

    def feature_locations(self):
        return collect_feature_locations(self.config.paths)

    def run(self):
        with self.path_manager:
            self.setup_paths()
            return self.run_with_paths()

    def run_with_paths(self):
        self.context = Context(self)
        self.load_hooks()
        self.load_step_definitions()

        # -- ENSURE: context.execute_steps() works in weird cases (hooks, ...)
        # self.setup_capture()
        # self.run_hook("before_all", self.context)

        # -- STEP: Parse all feature files (by using their file location).
        feature_locations = [filename for filename in self.feature_locations()
                             if not self.config.exclude(filename)]
        features = parse_features(feature_locations, language=self.config.lang)
        self.features.extend(features)

        # -- STEP: Run all features.
        stream_openers = self.config.outputs
        self.formatters = make_formatters(self.config, stream_openers)
        return self.run_model()


class BehaveRunner(Runner):
    """Раннер для запуска тестов.

    """

    def __init__(self, config, **kwargs):
        super().__init__(config)
        self.tempdir = kwargs.get('tempdir')
        self.step_template_data_types = kwargs.get('step_template_data_types', {})
        self.environment_filepath = kwargs.get('environment_filepath')

    @cached_property
    def tag_scenario_map(self):
        """Словарь соответствия тэгов и сценариев.
        """
        tags_map = {}

        for feature in self.features:
            for scenario in feature.scenarios:
                if scenario.main_tag in tags_map:
                    raise Exception(f'Дублирование основного тега сценария {scenario.main_tag}')
                else:
                    tags_map[scenario.main_tag] = scenario

        return tags_map

    def get_step_locations(self, feature=None) -> set:
        """Получает список всех путей до директорий с определением шагов.

        Returns:
            Расположения реализаций шагов.
        """
        step_locations = set()

        for path in self.config.paths:
            steps_dir_path = os.path.join(path, self.config.steps_dir)
            if os.path.isdir(steps_dir_path):
                step_locations.add(steps_dir_path)

        return step_locations

    def do_register_types(self):
        """Зарегистрировать пользовательские типы парсинга параметров из шагов."""
        ParseMatcher.parser_class = CaseSensitiveParser

        register_type(**self.step_template_data_types)

    def load_step_definitions(self, extra_step_paths=None, feature=None):
        """Переопределен для загрузки шагов из всех app'ов и регистрации пользовательских типов.

        Args:
            extra_step_paths: Дополнительные пути до директорий с определением шагов.
            feature: Объект класса Feature.
        """
        self.do_register_types()
        step_locations = set(extra_step_paths or [])
        step_locations.update(self.get_step_locations(feature=feature))

        steps_dir = os.path.join(self.base_dir, self.config.steps_dir)
        step_locations.add(steps_dir)

        load_step_modules(step_locations)

    def get_filtered_feature_locations(self):
        """Получить подходящие пути до feature-файлов."""
        return [
            filename
            for filename in self.feature_locations()
            if not self.config.exclude(filename)
        ]

    def sort_scenarios_by_given_tags(self):
        """Сортирует сценарии в соответствии с переданными тэгами."""

        def sort_key(s):
            """Ключ сортировки сценариев."""

            return (given_tags.index(s.main_tag), True) if s.main_tag in given_tags else (0, False)

        given_tags = self.config.tags.ands[0]
        features = []
        for tag in given_tags:
            feature = self.tag_scenario_map[tag].feature
            if feature not in features:
                self.features.remove(feature)
                feature.scenarios = sorted(feature.scenarios, key=sort_key)
                features.append(feature)

        self.features = features + self.features

    def save_features_steps_definition_cache(self):
        with open(self.config.optimize_features_steps_loading, 'w') as f:
            f.write(json.dumps({
                f.filename: {
                    scenario: list(definitions)
                    for scenario, definitions in f.steps_definitions_dirs.items()
                } for f in self.features
            }))
            
        tests_logger.info(
            f"Файл кэширования feature-файлов экспортирован: {self.config.optimize_features_steps_loading}"
        )

    def run_with_paths(self):
        """Изменен порядок загрузки определения шагов и парсинга feature файлов.

        """
        self.context = Context(self)
        self.load_hooks(filename=self.environment_filepath)

        feature_locations = self.get_filtered_feature_locations()
        features = parse_features(feature_locations, language=self.config.lang)
        self.features.extend(features)

        if self.config.in_tags_order:
            self.sort_scenarios_by_given_tags()

        self.load_step_definitions()

        if self.config.optimize_features_steps_loading:
            self.config.dry_run = True
            failed = self.run_model()
            self.save_features_steps_definition_cache()
        else:
            stream_openers = self.config.outputs
            self.formatters = make_formatters(self.config, stream_openers)
            failed = self.run_model()

        return failed


class BehaveTagsCachedRunner(BehaveRunner):
    """
    Раннер тестов с возможностью кэшировать в файлы пути до steps/ по тегам отдельных сценариев,
    для ускорения загрузки load_step_definitions.
    """

    def __init__(self, config, **kwargs):
        self.not_cached_tags = set()
        super().__init__(config, **kwargs)

    def run_with_paths(self):
        if self.config.clear_cached_step_locations:
            self.clear_cached_steps_dir()

        return super().run_with_paths()

    def get_step_locations(self, feature=None) -> set:
        """Получает пути до директорий с определением шагов из кэша.
        Если в кеше сценарий не найден, для него создается запись.

        Returns:
            Расположения реализаций шагов.
        """
        step_locations = set()

        # Определение тегов, относящихся к сценариям.
        scenario_tags = set(self.tag_scenario_map).intersection(sum(self.config.tags.ands, []))

        self.not_cached_tags = scenario_tags.difference(self.cached_scenario_tags)

        if self.not_cached_tags:
            step_locations = super().get_step_locations(feature=feature)
        else:
            for tag in scenario_tags:
                step_locations.update(self.cached_step_locations[tag])

        return step_locations

    def load_step_definitions(self, extra_step_paths=None, feature=None):
        super().load_step_definitions(extra_step_paths, feature)
        self.save_step_locations_in_cache()

    @property
    def cached_steps_dir(self):
        """Директория хранения кэшированных путей реализации шагов.

        """
        temp_dir = os.path.join(self.tempdir, 'cached_step_locations')
        if not os.path.exists(temp_dir):
            os.makedirs(temp_dir)

        return temp_dir

    def clear_cached_steps_dir(self):
        """Очищает кэшированные пути шагов сценариев.

        """
        if self.cached_scenario_tags:
            for f in glob(os.path.join(self.cached_steps_dir, '*')):
                os.remove(f)
            tests_logger.info('Кэшированные пути шагов сценариев очищены.')
        else:
            tests_logger.info('Кэшированные пути шагов не найдены.')

        exit(0)

    @cached_property
    def cached_scenario_tags(self):
        """Тэги кэшированных путей сценариев.

        """
        return set(os.listdir(self.cached_steps_dir))

    @cached_property
    def cached_step_locations(self):
        """Словарь соответствия тэгов и кэшированных путей реализации шагов.

        """
        step_locations = {}

        for tag in self.cached_scenario_tags:
            path = os.path.join(self.cached_steps_dir, tag)
            if os.path.isfile(path):
                with open(path) as f:
                    locations_str = f.read()
                locations = json.loads(locations_str)
                step_locations[tag] = locations

        return step_locations

    def save_step_locations_in_cache(self) -> None:
        """Сохраняет пути реализации шагов для ещё не закэшированных тегов."""

        for tag in self.not_cached_tags:
            scenario = self.tag_scenario_map[tag]
            scenario_step_locations = set()
            scenario_has_undefined_steps = True

            for n, step in enumerate(scenario.all_steps):
                try:
                    location = the_step_registry.find_match(step).location.dirname()
                except AttributeError:
                    if scenario.type == 'scenario_outline':
                        step = list(scenario.scenarios[0].all_steps)[n]
                        try:
                            location = the_step_registry.find_match(step).location.dirname()
                        except AttributeError:
                            break
                    else:
                        break
                scenario_step_locations.add(location)
            else:
                scenario_has_undefined_steps = False
                with open(os.path.join(self.cached_steps_dir, tag), 'w') as f:
                    tag_steps_locations_str = json.dumps(sorted(scenario_step_locations))
                    f.write(tag_steps_locations_str)

            if scenario_has_undefined_steps:
                tests_logger.warning(f'Scenario {tag} has undefined steps.')


class ParallelBehaveRunner(BehaveRunner):
    """Раннер для параллельного запуска тестов.

    """
    parallel_features_scenarios = None

    def __init__(self, config, **kwargs):
        self.features_steps_definitions_dirs = {}
        super().__init__(config, **kwargs)

    def should_run_feature(self, feature):
        """
        Проверка наличия сценариев для тестирования
        """
        return self.parallel_features_scenarios[hash(feature)]

    def pre_run_feature(self, feature):
        """Действия перед запуском feature"""
        if self.features_steps_definitions_dirs:
            self.load_step_definitions(feature=feature)

    def get_step_locations(self, feature=None) -> set:
        """
        Получить пути до директорий steps. Если есть закэшированные пути, использовать их.
        """
        feature_definitions = None
        step_locations = set()

        if feature and self.features_steps_definitions_dirs:
            try:
                feature_definitions = self.features_steps_definitions_dirs[feature.filename]
            except KeyError:
                # Если feature новая, или по какой-то причине не оказалось в файле,
                # попробуем найти путь по feature-файлу.
                step_locations.add(
                    os.path.join(os.path.dirname(feature.filename), self.config.steps_dir)
                )

            if feature_definitions:
                try:
                    for sc_hash in self.parallel_features_scenarios[hash(feature)]:
                        step_locations.update(feature_definitions[sc_hash])
                except KeyError as e:
                    tests_logger.warning(
                        f'Ошибка при получении путей к step_definitions сценария из '
                        f'{os.path.basename(feature.filename)}: {e}'
                    )
                    # добавим все пути относящиеся к feature
                    step_locations.update(itertools.chain(*feature_definitions.values()))
                    step_locations.add(
                        os.path.join(os.path.dirname(feature.filename), self.config.steps_dir)
                    )

        else:
            step_locations = super().get_step_locations(feature=feature)

        return step_locations

    def parallel_run_model(self, features=None, tags=None, results_list=None, **kwargs):
        """Запуск run_model в отдельном потоке и сбор результатов в results_list.

        Args:
            features: Список объектов Feature.
            tags: Список тэгов.
            results_list: Список хранения результатов выполнения потоков.
        """
        # Поиск группового тэга и запись его номера в group_number.
        group_tag = [tag for tag in tags if scenario_group_prefix in tag][0]
        self.config.group_number = group_tag.replace(scenario_group_prefix, '')

        # Изменим порядок следования feature для отдельного потока
        random.shuffle(features)

        failed = self.run_model(features)

        results_list.append(failed)

    def run_with_paths(self):
        """Переопределен для возможности запуска в несколько потоков.

        """
        self.context = Context(self)
        self.load_hooks(filename=self.environment_filepath)

        if self.config.optimize_features_steps_loading:
            try:
                with open(self.config.optimize_features_steps_loading) as f:
                    self.features_steps_definitions_dirs = {
                        feature: {
                            hash(step): set(definitions)
                            for step, definitions in scenarios.items()
                        }
                        for feature, scenarios in json.loads(f.read()).items()
                    }
            except Exception as e:
                tests_logger.warning(
                    f"Ошибка при попытке открыть файл {self.config.optimize_features_steps_loading}: {e}"
                )

        if not self.features_steps_definitions_dirs:
            self.load_step_definitions()

        feature_locations = self.get_filtered_feature_locations()
        features = parse_features(feature_locations, language=self.config.lang)
        self.features.extend(features)

        failed = self.run_parallel_processes()

        self.do_after_parallel_run()

        return failed

    def prepare_process_kwargs(self, manager):
        """Создаёт словарь именованных аргументов для отдельного процесса.
        В случае если у репортера определён атрибут process_kwarg_name, инициализируем переменную и добавляем в словарь.

        Args:
            manager: объект multiprocessing.Manager
        """
        data = {}

        for reporter in self.config.reporters:
            if hasattr(reporter, 'process_kwarg_name'):
                data[reporter.process_kwarg_name] = reporter.init_process_variable(manager)

        return data

    def do_after_parallel_run(self):
        """Вызывает метод after_parallel_run для каждого репортера у которого есть соответствующий метод.
        """
        for reporter in self.config.reporters:
            if hasattr(reporter, 'after_parallel_run'):
                reporter.after_parallel_run()

    def prepare_parallel_feature_scenarios_map(self, manager, features):
        """Подготовить DictProxy карту сценариев по указанным feature-файлам.

        Args:
            manager: объект multiprocessing.Manager
            features: список объектов Feature

        Returns:
            DictProxy с ключами feature и значениями ListProxy --
             списком идентификаторов сценариев которые нужно выполнить.
        """
        self.parallel_features_scenarios = manager.dict()

        for feature in features:
            sc_keys = list(feature.scenarios_map.keys())
            random.shuffle(sc_keys)

            self.parallel_features_scenarios[hash(feature)] = manager.list(sc_keys)

        return self.parallel_features_scenarios

    def run_parallel_processes(self):
        """Запуск тестов в параллельных процессах.

        """
        groups, used_features = self.prepare_groups_and_features(
            self.config.parallel_count,
        )

        manager = multiprocessing.Manager()
        failed_results_list = manager.list()

        process_kwargs = self.prepare_process_kwargs(manager)
        process_kwargs['parallel_features_scenarios'] = self.prepare_parallel_feature_scenarios_map(
            manager,
            used_features,
        )

        jobs = []

        for group_tag in groups:
            process = multiprocessing.Process(
                target=self.parallel_run_model,
                args=(
                    used_features,
                    [group_tag],
                    failed_results_list,
                ),
                kwargs=process_kwargs,
            )
            jobs.append(process)

        start_time = datetime.datetime.now()

        for j in jobs:
            j.start()

        for j in jobs:
            j.join()

        if self.config.junit:
            self.combine_junit_reports(start_time)

        # Если хотя бы один результат True - значит было падение тестов
        return any(failed_results_list)

    def prepare_groups_and_features(self, groups_count=1):
        """Формирует группы, в зависимости от количества сценариев и указанного числа групп.
        Так же формирует список feature-файлов которые будут запускаться в зависимости от тегов.

        Args:
            groups_count: Количество групп.

        Returns:
            Список тегов групп, список объектов Feature
        """
        used_features = set()
        groups = []
        scenarios_count = 0

        for feature in self.features:
            if 'skip' in feature.tags:
                continue

            # сформируем карту соответствия идентификатора сценария и объекта сценария
            feature.scenarios_map = {}

            for sc in feature:
                if sc.should_run_with_tags(self.config.tags):
                    used_features.add(sc.feature)

                    if sc.type == 'scenario_outline':
                        for sub_sc in sc.scenarios:
                            feature.scenarios_map[hash(f'{sub_sc.name}{sub_sc.location}')] = sub_sc
                    else:
                        feature.scenarios_map[hash(f'{sc.name}{sc.location}')] = sc

            scenarios_count += len(feature.scenarios_map.keys())

        limit = (
            groups_count
            if scenarios_count > groups_count else
            scenarios_count
        )

        for i in range(limit):
            # фиктивный тег для группы
            group_tag = f'{scenario_group_prefix}{i}'
            groups.append(group_tag)

        return groups, list(used_features)

    def combine_junit_reports(self, start_time):
        """Объединяет части junit-отчётов полученные при выполнении параллельных процессов.

        Args:
            start_time: Время запуска процессов.

        """
        reports_combiner = JunitReportsCombiner(
            tests_execution_start_time=start_time,
            reports_dir_path=self.config.junit_directory,
            rerun=getattr(self.config, 'rerun', False),
        )
        reports_combiner.join_reports()

        if self.config.collect_top_scenarios:
            reports_combiner.export_top_scenarios_by_time()



"""
Capture output (stdout, stderr), logs, etc.
"""

import sys
from contextlib import (
    contextmanager,
)
from io import (
    StringIO,
)

from behave_bo.log_capture import (
    LoggingCapture,
)
from behave_bo.textutil import (
    text as _text,
)


def add_text_to(value, more_text, separator="\n"):
    if more_text:
        if value:
            if separator and not value.endswith(separator):
                value += separator
            value += more_text
        else:
            value = more_text
    return value


class Captured:
    """Stores and aggregates captured output data."""
    empty = ""
    linesep = "\n"

    def __init__(self, stdout=None, stderr=None, log_output=None):
        self.stdout = stdout or self.empty
        self.stderr = stderr or self.empty
        self.log_output = log_output or self.empty

    def reset(self):
        self.stdout = self.empty
        self.stderr = self.empty
        self.log_output = self.empty

    # -- PYTHON2:
    def __bool__(self):
        return bool(self.stdout or self.stderr or self.log_output)

    @property
    def output(self):
        """Makes a simple report of the captured data by concatenating
        all parts.
        """
        output_text = self.stdout
        output_text = add_text_to(output_text, self.stderr)
        output_text = add_text_to(output_text, self.log_output)
        return output_text

    def add(self, captured):
        """Adds/appends captured output data to this object.

        :param captured:    Captured object whose data should be added.
        :return: self, to allow daisy-chaining (if needed).
        """
        assert isinstance(captured, Captured)
        self.stdout = add_text_to(self.stdout, captured.stdout, self.linesep)
        self.stderr = add_text_to(self.stderr, captured.stderr, self.linesep)
        self.log_output = add_text_to(self.log_output, captured.log_output,
                                      self.linesep)
        return self

    def make_report(self):
        """Makes a detailled report of the captured output data.

        :returns: Report as string.
        """
        report_parts = []
        if self.stdout:
            parts = ["Captured stdout:", _text(self.stdout).rstrip(), ""]
            report_parts.extend(parts)
        if self.stderr:
            parts = ["Captured stderr:", _text(self.stderr).rstrip(), ""]
            report_parts.extend(parts)
        if self.log_output:
            parts = ["Captured logging:", _text(self.log_output)]
            report_parts.extend(parts)
        return self.linesep.join(report_parts).strip()

    def __add__(self, other):
        """Supports incremental add::

            captured1 = Captured("Hello")
            captured2 = Captured("World")
            captured3 = captured1 + captured2
            assert captured3.stdout == "Hello\nWorld"
        """
        new_data = Captured(self.stdout, self.stderr, self.log_output)
        return new_data.add(other)

    def __iadd__(self, other):
        """Supports incremental add::

            captured1 = Captured("Hello")
            captured2 = Captured("World")
            captured1 += captured2
            assert captured1.stdout == "Hello\nWorld"
        """
        return self.add(other)


class CaptureController:
    """Simplifies the lifecycle to capture output from various sources."""
    def __init__(self, config):
        self.config = config
        self.stdout_capture = None
        self.stderr_capture = None
        self.log_capture = None
        self.old_stdout = None
        self.old_stderr = None

    @property
    def captured(self):
        """Provides access of the captured output data.

        :return: Object that stores the captured output parts (as Captured).
        """
        stdout = None
        stderr = None
        log_out = None
        if self.config.stdout_capture and self.stdout_capture:
            stdout = _text(self.stdout_capture.getvalue())
        if self.config.stderr_capture and self.stderr_capture:
            stderr = _text(self.stderr_capture.getvalue())
        if self.config.log_capture and self.log_capture:
            log_out = _text(self.log_capture.getvalue())
        return Captured(stdout, stderr, log_out)

    def setup_capture(self, context):
        assert context is not None
        if self.config.stdout_capture:
            self.stdout_capture = StringIO()
            context.stdout_capture = self.stdout_capture

        if self.config.stderr_capture:
            self.stderr_capture = StringIO()
            context.stderr_capture = self.stderr_capture

        if self.config.log_capture:
            self.log_capture = LoggingCapture(self.config)
            self.log_capture.inveigle()
            context.log_capture = self.log_capture

    def start_capture(self):
        if self.config.stdout_capture:
            # -- REPLACE ONLY: In non-capturing mode.
            if not self.old_stdout:
                self.old_stdout = sys.stdout
                sys.stdout = self.stdout_capture
            assert sys.stdout is self.stdout_capture

        if self.config.stderr_capture:
            # -- REPLACE ONLY: In non-capturing mode.
            if not self.old_stderr:
                self.old_stderr = sys.stderr
                sys.stderr = self.stderr_capture
            assert sys.stderr is self.stderr_capture

    def stop_capture(self):
        if self.config.stdout_capture:
            # -- RESTORE ONLY: In capturing mode.
            if self.old_stdout:
                sys.stdout = self.old_stdout
                self.old_stdout = None
            assert sys.stdout is not self.stdout_capture

        if self.config.stderr_capture:
            # -- RESTORE ONLY: In capturing mode.
            if self.old_stderr:
                sys.stderr = self.old_stderr
                self.old_stderr = None
            assert sys.stderr is not self.stderr_capture

    def teardown_capture(self):
        if self.config.log_capture:
            self.log_capture.abandon()

    def make_capture_report(self):
        """Combine collected output and return as string."""
        return self.captured.make_report()
        # report = u""
        # if self.config.stdout_capture and self.stdout_capture:
        #     output = self.stdout_capture.getvalue()
        #     if output:
        #         output = _text(output)
        #         report += u"\nCaptured stdout:\n" + output
        # if self.config.stderr_capture and self.stderr_capture:
        #     output = self.stderr_capture.getvalue()
        #     if output:
        #         output = _text(output)
        #         report += u"\nCaptured stderr:\n" + output
        # if self.config.log_capture and self.log_capture:
        #     output = self.log_capture.getvalue()
        #     if output:
        #         output = _text(output)
        #         report += u"\nCaptured logging:\n" + output
        # return report

# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# -----------------------------------------------------------------------------
@contextmanager
def capture_output(controller, enabled=True):
    """Provides a context manager that starts capturing output

    .. code-block::

        with capture_output(capture_controller):
            ... # Do something
    """
    if enabled:
        try:
            controller.start_capture()
            yield
        finally:
            controller.stop_capture()
    else:
        # -- CAPTURING OUTPUT is disabled.
        # Needed to prevent recursive captures with context.execute_steps()
        yield



import os
import shutil
import tempfile
import uuid
from contextlib import (
    contextmanager,
)

from django.apps import (
    apps,
)
from django.conf import (
    settings,
)
from django.test.utils import (
    override_settings,
)


def get_features_paths():
    """Возвращает список путей до директорий features в app'ах, если существуют."""

    paths = []
    for app in apps.get_app_configs():
        features_dir = os.path.join(app.path, 'features')
        if os.path.isdir(features_dir):
            paths.append(features_dir)

    return paths


@contextmanager
def override_dirs(scenario):
    """Контекстный менеджер для создания уникальных директорий под временные файлы и downloads для выполняемого сценария

    Args:
        scenario: объект класса Scenario

    """
    prefix = scenario.main_tag if scenario.tags else uuid.uuid4().hex[:6]

    old_tempdir = settings.TEMPDIR
    old_downloads_dir = settings.DOWNLOADS_DIR

    new_tempdir = f"{old_tempdir}/{prefix}__{scenario.line}"
    new_downloads_dir = f"{old_downloads_dir}/{prefix}__{scenario.line}"

    if os.path.isdir(new_tempdir):
        shutil.rmtree(new_tempdir, ignore_errors=True)

    tempfile.tempdir = new_tempdir

    if not os.path.isdir(tempfile.gettempdir()):
        os.makedirs(tempfile.gettempdir())

    if os.path.isdir(new_downloads_dir):
        shutil.rmtree(new_downloads_dir, ignore_errors=True)

    if not os.path.isdir(new_downloads_dir):
        os.makedirs(new_downloads_dir)

    try:
        with override_settings(TEMPDIR=new_tempdir, DOWNLOADS_DIR=new_downloads_dir):
            yield
    finally:
        tempfile.tempdir = old_tempdir



import os
import traceback
from functools import (
    wraps,
)

from behave_bo.loggers import (
    tests_logger,
)
from coverage import (
    Coverage,
    CoverageException,
)


def coverage_switch_context(func):
    """
    Декоратор для переключения динамического контекста при анализе покрытия кода.
    Прокидывает в отчёт покрытия информацию о файле из которого был отправлен запрос.
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):

        if self._context.config.measure_requests_coverage:
            coverage = Coverage.current()
            frame = traceback.extract_stack()[-2]  # фрейм с файлом и функцией отправки запроса
            basename = os.path.basename(frame.filename)

            try:
                if coverage:
                    tests_logger.info(f'switch context! {basename}')
                    coverage.switch_context(basename)
            except CoverageException as e:
                tests_logger.info(str(e))

        return func(self, *args, **kwargs)

    return wrapper



# UNUSED: import sys
# UNUSED: import codecs
import os.path

from behave_bo.textutil import (
    ensure_stream_with_encoder as _ensure_stream_with_encoder,
    select_best_encoding,
)


class StreamOpener:
    """Provides a transport vehicle to open the formatter output stream
    when the formatter needs it.
    In addition, it provides the formatter with more control:

      * when a stream is opened
      * if a stream is opened at all
      * the name (filename/dirname) of the output stream
      * let it decide if directory mode is used instead of file mode
    """
    # FORMER: default_encoding = "UTF-8"
    default_encoding = select_best_encoding()

    def __init__(self, filename=None, stream=None, encoding=None):
        if not encoding:
            encoding = self.default_encoding
        if stream:
            stream = self.ensure_stream_with_encoder(stream, encoding)
        self.name = filename
        self.stream = stream
        self.encoding = encoding
        self.should_close_stream = not stream   # Only for not pre-opened ones.

    @staticmethod
    def ensure_dir_exists(directory):
        if directory and not os.path.isdir(directory):
            os.makedirs(directory)

    @classmethod
    def ensure_stream_with_encoder(cls, stream, encoding=None):
        return _ensure_stream_with_encoder(stream, encoding)

    def open(self):
        if not self.stream or self.stream.closed:
            self.ensure_dir_exists(os.path.dirname(self.name))
            stream = open(self.name, "w")
            # stream = codecs.open(self.name, "w", encoding=self.encoding)
            stream = self.ensure_stream_with_encoder(stream, self.encoding)
            self.stream = stream  # -- Keep stream for house-keeping.
            self.should_close_stream = True
            assert self.should_close_stream
        return self.stream

    def close(self):
        """
        Close the stream, if it was opened by this stream_opener.
        Skip closing for sys.stdout and pre-opened streams.
        :return: True, if stream was closed.
        """
        closed = False
        if self.stream and self.should_close_stream:
            closed = getattr(self.stream, "closed", False)
            if not closed:
                self.stream.close()
                closed = True
            self.stream = None
        return closed


class Formatter:
    """
    Base class for all formatter classes.
    A formatter is an extension point (variation point) for the runner logic.
    A formatter is called while processing model elements.

    Processing Logic (simplified, without ScenarioOutline and skip logic)::

        for feature in runner.features:
            formatter = make_formatters(...)
            formatter.uri(feature.filename)
            formatter.feature(feature)
            for scenario in feature.scenarios:
                formatter.scenario(scenario)
                for step in scenario.all_steps:
                    formatter.step(step)
                    step_match = step_registry.find_match(step)
                    formatter.match(step_match)
                    if step_match:
                        step_match.run()
                    else:
                        step.status = Status.undefined
                    formatter.result(step.status)
            formatter.eof() # -- FEATURE-END
        formatter.close()
    """
    name = None
    description = None

    def __init__(self, stream_opener, config):
        self.stream_opener = stream_opener
        self.stream = stream_opener.stream
        self.config = config

    @property
    def stdout_mode(self):
        return not self.stream_opener.name

    def open(self):
        """
        Ensure that the output stream is open.
        Triggers the stream opener protocol (if necessary).

        :return: Output stream to use (just opened or already open).
        """
        if not self.stream:
            self.stream = self.stream_opener.open()
        return self.stream

    def uri(self, uri):
        """Called before processing a file (normally a feature file).

        :param uri:  URI or filename (as string).
        """
        pass

    def feature(self, feature):
        """Called before a feature is executed.

        :param feature:  Feature object (as :class:`behave_bo.model.Feature`)
        """
        pass

    def background(self, background):
        """Called when a (Feature) Background is provided.
        Called after :method:`feature()` is called.
        Called before processing any scenarios or scenario outlines.

        :param background:  Background object (as :class:`behave_bo.model.Background`)
        """
        pass

    def scenario(self, scenario):
        """Called before a scenario is executed (or ScenarioOutline scenarios).

        :param scenario:  Scenario object (as :class:`behave_bo.model.Scenario`)
        """
        pass

    def step(self, step):
        """Called before a step is executed (and matched).
        NOTE: Normally called before scenario is executed for all its steps.

        :param step: Step object (as :class:`behave_bo.model.Step`)
        """
        pass

    def match(self, match):
        """Called when a step was matched against its step implementation.

        :param match:  Registered step (as Match), undefined step (as NoMatch).
        """
        pass

    def result(self, step):
        """Called after processing a step (when the step result is known).

        :param step:  Step object with result (after being executed/skipped).
        """
        pass

    def eof(self):
        """Called after processing a feature (or a feature file)."""
        pass

    def close(self):
        """Called before the formatter is no longer used
        (stream/io compatibility).
        """
        self.close_stream()

    def close_stream(self):
        """Close the stream, but only if this is needed.
        This step is skipped if the stream is sys.stdout.
        """
        if self.stream:
            # -- DELEGATE STREAM-CLOSING: To stream_opener
            assert self.stream is self.stream_opener.stream
            self.stream_opener.close()
        self.stream = None      # -- MARK CLOSED.



class TagExpression:
    """
    Tag expression, as logical boolean expression, to select
    (include or exclude) model elements.

    BOOLEAN LOGIC := (or_expr1) and (or_expr2) and ...
    with or_exprN := [not] tag1 or [not] tag2 or ...
    """

    def __init__(self, tag_expressions):
        self.ands = []
        self.limits = {}

        for expr in tag_expressions:
            self.store_and_extract_limits(self.normalized_tags_from_or(expr))

    @staticmethod
    def normalize_tag(tag):
        """
        Normalize a tag for a tag expression:

          * strip whitespace
          * strip '@' char
          * convert '~' (tilde) into '-' (minus sign)

        :param tag:  Tag (as string).
        :return: Normalized tag (as string).
        """
        tag = tag.strip()
        if tag.startswith('@'):
            tag = tag[1:]
        elif tag.startswith('-@') or tag.startswith('~@'):
            tag = '-' + tag[2:]
        elif tag.startswith('~'):
            tag = '-' + tag[1:]
        return tag

    @classmethod
    def normalized_tags_from_or(cls, expr):
        """
        Normalizes all tags in an OR expression (and return it as list).

        :param expr:  OR expression to normalize and split (as string).
        :return: Generator of normalized tags (as string)
        """
        for tag in expr.strip().split(','):
            yield cls.normalize_tag(tag)

    def store_and_extract_limits(self, tags):
        tags_with_negation = []

        for tag in tags:
            negated = tag.startswith('-')
            tag = tag.split(':')
            tag_with_negation = tag.pop(0)
            tags_with_negation.append(tag_with_negation)

            if tag:
                limit = int(tag[0])
                if negated:
                    tag_without_negation = tag_with_negation[1:]
                else:
                    tag_without_negation = tag_with_negation
                limited = tag_without_negation in self.limits
                if limited and self.limits[tag_without_negation] != limit:
                    msg = "Inconsistent tag limits for {0}: {1:d} and {2:d}"
                    msg = msg.format(tag_without_negation,
                                     self.limits[tag_without_negation], limit)
                    raise Exception(msg)
                self.limits[tag_without_negation] = limit

        if tags_with_negation:
            self.ands.append(tags_with_negation)

    def check(self, tags):
        """
        Checks if this tag expression matches the tags of a model element.

        :param tags:  List of tags of a model element.
        :return: True, if tag expression matches. False, otherwise.
        """
        if not self.ands:
            return True

        element_tags = set(tags)

        def test_tag(xtag):
            if xtag.startswith('-'): # -- or xtag.startswith('~'):
                return xtag[1:] not in element_tags
            return xtag in element_tags

        # -- EVALUATE: (or_expr1) and (or_expr2) and ...
        return all(any(test_tag(xtag) for xtag in ors)  for ors in self.ands)

    def __len__(self):
        return len(self.ands)

    def __str__(self):
        """Conversion back into string that represents this tag expression."""
        and_parts = []
        for or_terms in self.ands:
            and_parts.append(",".join(or_terms))
        return " ".join(and_parts)



import sys

from behave_bo import (
    __version__,
)
from behave_bo.configuration import (
    ConfigError,
    Configuration,
)
from behave_bo.parser import (
    ParserError,
)
from behave_bo.runner import (
    Runner,
)
from behave_bo.runner_util import (
    FileNotFoundError,
    InvalidFileLocationError,
    InvalidFilenameError,
    print_undefined_step_snippets,
    reset_runtime,
)
from behave_bo.textutil import (
    compute_words_maxsize,
    text as _text,
)


TAG_HELP = """
Scenarios inherit tags declared on the Feature level. The simplest
TAG_EXPRESSION is simply a tag::

    --tags @dev

You may even leave off the "@" - behave_bo doesn't mind.

When a tag in a tag expression starts with a ~, this represents boolean NOT::

    --tags ~@dev

A tag expression can have several tags separated by a comma, which represents
logical OR::

    --tags @dev,@wip

The --tags option can be specified several times, and this represents logical
AND, for instance this represents the boolean expression
"(@foo or not @bar) and @zap"::

    --tags @foo,~@bar --tags @zap.

Beware that if you want to use several negative tags to exclude several tags
you have to use logical AND::

    --tags ~@fixme --tags ~@buggy.
""".strip()


def run_behave(config, runner_class=None, runner_kwargs=None):
    """Run behave_bo with configuration.

    :param config:          Configuration object for behave_bo.
    :param runner_class:    Runner class to use or none (use Runner class).
    :param runner_kwargs:   Runner class init kwargs
    :return:    0, if successful. Non-zero on failure.

    .. note:: BEST EFFORT, not intended for multi-threaded usage.
    """
    # pylint: disable=too-many-branches, too-many-statements, too-many-return-statements
    if runner_class is None:
        runner_class = Runner

    if runner_kwargs is None:
        runner_kwargs = {}

    if config.version:
        print("behave " + __version__)
        return 0

    if config.tags_help:
        print(TAG_HELP)
        return 0

    if config.lang_list:
        from behave_bo.i18n import (
            languages,
        )
        stdout = sys.stdout
        iso_codes = languages.keys()
        print("Languages available:")
        for iso_code in sorted(iso_codes):
            native = languages[iso_code]["native"][0]
            name = languages[iso_code]["name"][0]
            print(f"{iso_code}: {native} / {name}", file=stdout)
        return 0

    if config.lang_help:
        from behave_bo.i18n import (
            languages,
        )
        if config.lang_help not in languages:
            print("%s is not a recognised language: try --lang-list" % \
                  config.lang_help)
            return 1
        trans = languages[config.lang_help]
        print("Translations for {} / {}".format(trans["name"][0],
                                                trans["native"][0]))
        for kw in trans:
            if kw in "name native".split():
                continue
            print("%16s: %s" % (kw.title().replace("_", " "),
                                ", ".join(w for w in trans[kw] if w != "*")))
        return 0

    if not config.format:
        config.format = [config.default_format]
    elif config.format and "format" in config.defaults:
        # -- CASE: Formatter are specified in behave_bo configuration file.
        #    Check if formatter are provided on command-line, too.
        if len(config.format) == len(config.defaults["format"]):
            # -- NO FORMATTER on command-line: Add default formatter.
            config.format.append(config.default_format)
    if "help" in config.format:
        print_formatters("Available formatters:")
        return 0

    if len(config.outputs) > len(config.format):
        print("CONFIG-ERROR: More outfiles (%d) than formatters (%d)." % \
              (len(config.outputs), len(config.format)))
        return 1

    # -- MAIN PART:
    failed = True
    try:
        reset_runtime()
        runner = runner_class(config, **runner_kwargs)
        failed = runner.run()
    except ParserError as e:
        print("ParserError: %s" % e)
    except ConfigError as e:
        print("ConfigError: %s" % e)
    except FileNotFoundError as e:
        print("FileNotFoundError: %s" % e)
    except InvalidFileLocationError as e:
        print("InvalidFileLocationError: %s" % e)
    except InvalidFilenameError as e:
        print("InvalidFilenameError: %s" % e)
    except Exception as e:
        # -- DIAGNOSTICS:
        text = _text(e)
        print(f"Exception {e.__class__.__name__}: {text}")
        raise

    if config.show_snippets and runner.undefined_steps:
        print_undefined_step_snippets(runner.undefined_steps,
                                      colored=config.color)

    return_code = 0
    if failed:
        return_code = 1
    return return_code


def print_formatters(title=None, stream=None):
    """Prints the list of available formatters and their description.

    :param title:   Optional title (as string).
    :param stream:  Optional, output stream to use (default: sys.stdout).
    """
    from operator import (
        itemgetter,
    )

    from behave_bo.formatter._registry import (
        format_items,
    )

    if stream is None:
        stream = sys.stdout
    if title:
        stream.write("%s\n" % title)

    format_items = sorted(format_items(resolved=True), key=itemgetter(0))
    format_names = [item[0] for item in format_items]
    column_size = compute_words_maxsize(format_names)
    schema = "  %-" + _text(column_size) + "s  %s\n"
    for name, formatter_class in format_items:
        formatter_description = getattr(formatter_class, "description", "")
        stream.write(schema % (name, formatter_description))


def main(args=None):
    """Main function to run behave_bo (as program).

    :param args:    Command-line args (or string) to use.
    :return: 0, if successful. Non-zero, in case of errors/failures.
    """
    config = Configuration(args)
    return run_behave(config)


if __name__ == "__main__":
    # -- EXAMPLE: main("--version")
    sys.exit(main())



from django.core.management.base import (
    BaseCommand,
)

from behave_bo.reporter.junit import (
    JunitReportsCombiner,
)


class Command(BaseCommand):
    help = "Обрабатывает junit xml-файлы, формирует общий html-отчёт выполнения автотестов."

    def add_arguments(self, parser):
        parser.add_argument('junit_directory', help='путь до директории xml-файлов отчётов junit', type=str)
        parser.add_argument('report_directory', help='директория в которой будет сформирован html-отчёт', type=str)

    def handle(self, *args, **options):
        reports_combiner = JunitReportsCombiner(
            reports_dir_path=options['junit_directory'],
            html_report_destination_path=options['report_directory'],
        )
        report_path = reports_combiner.export_html_report()
        self.stdout.write(
            self.style.SUCCESS(f'Html-отчёт результатов автотестов выгружен в файл {report_path}')
        )



"""
Provides a summary after each test run.
"""

import os
import sys

from behave_bo.formatter.base import (
    StreamOpener,
)
from behave_bo.model import (
    ScenarioOutline,
)
from behave_bo.model_core import (
    Status,
)
from behave_bo.reporter.base import (
    Reporter,
)


# -- DISABLED: optional_steps = ('untested', 'undefined')
optional_steps = (Status.untested,) # MAYBE: Status.undefined
status_order = (Status.passed, Status.failed, Status.skipped,
                Status.undefined, Status.untested)


def format_summary(statement_type, summary):
    parts = []
    for status in status_order:
        if status.name not in summary:
            continue
        counts = summary[status.name]
        if status in optional_steps and counts == 0:
            # -- SHOW-ONLY: For relevant counts, suppress: untested items, etc.
            continue

        if not parts:
            # -- FIRST ITEM: Add statement_type to counter.
            label = statement_type
            if counts != 1:
                label += 's'
            part = "%d %s %s" % (counts, label, status.name)
        else:
            part = "%d %s" % (counts, status.name)
        parts.append(part)
    return ", ".join(parts) + "\n"


class SummaryReporter(Reporter):
    show_failed_scenarios = True
    output_stream_name = "stdout"

    def __init__(self, config):
        super().__init__(config)
        stream = getattr(sys, self.output_stream_name, sys.stderr)
        self.stream = StreamOpener.ensure_stream_with_encoder(stream)
        self.feature_summary = {Status.passed.name: 0, Status.failed.name: 0,
                                Status.skipped.name: 0, Status.untested.name: 0}
        self.scenario_summary = {Status.passed.name: 0, Status.failed.name: 0,
                                 Status.skipped.name: 0, Status.untested.name: 0}
        self.step_summary = {Status.passed.name: 0, Status.failed.name: 0,
                             Status.skipped.name: 0, Status.untested.name: 0,
                             Status.undefined.name: 0}
        self.duration = 0.0
        self.failed_scenarios = []

    def feature(self, feature):
        self.feature_summary[feature.status.name] += 1
        self.duration += feature.duration
        for scenario in feature:
            if isinstance(scenario, ScenarioOutline):
                self.process_scenario_outline(scenario)
            else:
                self.process_scenario(scenario)

    def end(self):
        # -- SHOW FAILED SCENARIOS (optional):
        if self.show_failed_scenarios and self.failed_scenarios:
            self.stream.write("\nFailing scenarios:\n")
            for scenario in self.failed_scenarios:
                self.stream.write("  {}  {}\n".format(
                    scenario.location, scenario.name))
            self.stream.write("\n")

        # -- SHOW SUMMARY COUNTS:
        self.stream.write(format_summary("feature", self.feature_summary))
        self.stream.write(format_summary("scenario", self.scenario_summary))
        self.stream.write(format_summary("step", self.step_summary))
        minutes, sec = (int(self.duration / 60.0), self.duration % 60)
        self.stream.write(f'Took {minutes}m{sec:2.3f}s (pid {os.getpid()})\n')

    def process_scenario(self, scenario):
        if scenario.status == Status.failed:
            self.failed_scenarios.append(scenario)

        self.scenario_summary[scenario.status.name] += 1
        for step in scenario:
            self.step_summary[step.status.name] += 1

    def process_scenario_outline(self, scenario_outline):
        for scenario in scenario_outline.scenarios:
            self.process_scenario(scenario)



"""
This module provides the abstract base classes and core concepts
for the model elements in behave_bo.
"""

import os.path
import sys

from behave_bo.capture import (
    Captured,
)
from behave_bo.textutil import (
    text as _text,
)

from enum import (
    Enum,
)


PLATFORM_WIN = sys.platform.startswith("win")
def posixpath_normalize(path):
    return path.replace("\\", "/")


# -----------------------------------------------------------------------------
# GENERIC MODEL CLASSES:
# -----------------------------------------------------------------------------
class Status(Enum):
    """Provides the (test-run) status of a model element.
    Features and Scenarios use: untested, skipped, passed, failed.
    Steps may use all enum-values.

    Enum values:
    * untested (initial state):

        Defines the initial state before a test-run.
        Sometimes used to indicate that the model element was not executed
        during a test run.

    * skipped:

        A model element is skipped because it should not run.
        This is caused by filtering mechanisms, like tags, active-tags,
        file-location arg, select-by-name, etc.

    * passed: A model element was executed and passed (without failures).
    * failed: Failures occurred while executing it.
    * undefined: Used for undefined-steps (no step implementation was found).
    * executing: Marks the steps during execution (used in a formatter)

    .. versionadded:: 1.2.6
        Superceeds string-based status values.
    """
    untested = 0
    skipped = 1
    passed = 2
    failed = 3
    undefined = 4
    executing = 5

    def __eq__(self, other):
        """Comparison operator equals-to other value.
        Supports other enum-values and string (for backward compatibility).

        EXAMPLES::

            status = Status.passed
            assert status == Status.passed
            assert status == "passed"
            assert status != "failed"

        :param other:   Other value to compare (enum-value, string).
        :return: True, if both values are equal. False, otherwise.
        """
        if isinstance(other, str):
            # -- CONVENIENCE: Compare with string-name (backward-compatible)
            return self.name == other
        return super().__eq__(other)

    @classmethod
    def from_name(cls, name):
        """Select enumeration value by using its name.

        :param name:    Name as key to the enum value (as string).
        :return: Enum value (instance)
        :raises: LookupError, if status name is unknown.
        """
        # pylint: disable=no-member
        enum_value = cls.__members__.get(name, None)
        if enum_value is None:
            known_names = ", ".join(cls.__members__.keys())
            raise LookupError(f"{name} (expected: {known_names})")
        return enum_value


class Argument:
    """An argument found in a *feature file* step name and extracted using
    step decorator `parameters`_.

    The attributes are:

    .. attribute:: original

       The actual text matched in the step name.

    .. attribute:: value

       The potentially type-converted value of the argument.

    .. attribute:: name

       The name of the argument. This will be None if the parameter is
       anonymous.

    .. attribute:: start

       The start index in the step name of the argument. Used for display.

    .. attribute:: end

       The end index in the step name of the argument. Used for display.
    """
    def __init__(self, start, end, original, value, name=None):
        self.start = start
        self.end = end
        self.original = original
        self.value = value
        self.name = name


# @total_ordering
# class FileLocation(unicode):
class FileLocation:
    """
    Provides a value object for file location objects.
    A file location consists of:

      * filename
      * line (number), optional

    LOCATION SCHEMA:
      * "{filename}:{line}" or
      * "{filename}" (if line number is not present)
    """
    __pychecker__ = "missingattrs=line"     # -- Ignore warnings for 'line'.

    def __init__(self, filename, line=None):
        if PLATFORM_WIN:
            filename = posixpath_normalize(filename)
        self.filename = filename
        self.line = line

    def get(self):
        return self.filename

    def abspath(self):
        return os.path.abspath(self.filename)

    def basename(self):
        return os.path.basename(self.filename)

    def dirname(self):
        return os.path.dirname(self.filename)

    def relpath(self, start=os.curdir):
        """Compute relative path for start to filename.

        :param start: Base path or start directory (default=current dir).
        :return: Relative path from start to filename
        """
        return os.path.relpath(self.filename, start)

    def exists(self):
        return os.path.exists(self.filename)

    def _line_lessthan(self, other_line):
        if self.line is None:
            # return not (other_line is None)
            return other_line is not None
        elif other_line is None:
            return False
        else:
            return self.line < other_line

    def __eq__(self, other):
        if isinstance(other, FileLocation):
            return self.filename == other.filename and self.line == other.line
        elif isinstance(other, str):
            return self.filename == other
        else:
            raise TypeError("Cannot compare FileLocation with %s:%s" % \
                            (type(other), other))

    def __ne__(self, other):
        # return not self == other    # pylint: disable=unneeded-not
        return not self.__eq__(other)

    def __lt__(self, other):
        if isinstance(other, FileLocation):
            if self.filename < other.filename:
                return True
            elif self.filename > other.filename:
                return False
            else:
                assert self.filename == other.filename
                return self._line_lessthan(other.line)

        elif isinstance(other, str):
            return self.filename < other
        else:
            raise TypeError("Cannot compare FileLocation with %s:%s" % \
                            (type(other), other))

    def __le__(self, other):
        # -- SEE ALSO: python2.7, functools.total_ordering
        # return not other < self     # pylint unneeded-not
        return other >= self

    def __gt__(self, other):
        # -- SEE ALSO: python2.7, functools.total_ordering
        if isinstance(other, FileLocation):
            return other < self
        else:
            return self.filename > other

    def __ge__(self, other):
        # -- SEE ALSO: python2.7, functools.total_ordering
        # return not self < other
        return not self.__lt__(other)

    def __repr__(self):
        return '<FileLocation: filename="%s", line=%s>' % \
               (self.filename, self.line)

    def __str__(self):
        filename = self.filename
        if isinstance(filename, bytes):
            filename = _text(filename, "utf-8")
        if self.line is None:
            return filename
        return "%s:%d" % (filename, self.line)

    @classmethod
    def for_function(cls, func, curdir=None):
        """Extracts the location information from the function and builds
        the location string (schema: "{source_filename}:{line_number}").

        :param func: Function whose location should be determined.
        :return: FileLocation object
        """
        func = unwrap_function(func)
        function_code = func.__code__
        filename = function_code.co_filename
        line_number = function_code.co_firstlineno

        curdir = curdir or os.getcwd()
        try:
            filename = os.path.relpath(filename, curdir)
        except ValueError:
            # WINDOWS-SPECIFIC (#599):
            # If a step-function comes from a different disk drive,
            # a relative path will fail: Keep the absolute path.
            pass
        return cls(filename, line_number)


# -----------------------------------------------------------------------------
# ABSTRACT MODEL CLASSES (and concepts):
# -----------------------------------------------------------------------------
class BasicStatement:
    def __init__(self, filename, line, keyword, name):
        filename = filename or '<string>'
        filename = os.path.relpath(filename, os.getcwd())   # -- NEEDS: abspath?
        self.location = FileLocation(filename, line)
        assert isinstance(keyword, str)
        assert isinstance(name, str)
        self.keyword = keyword
        self.name = name
        # -- SINCE: 1.2.6
        self.captured = Captured()
        # -- ERROR CONTEXT INFO:
        self.exception = None
        self.exc_traceback = None
        self.error_message = None

    @property
    def filename(self):
        # return os.path.abspath(self.location.filename)
        return self.location.filename

    @property
    def line(self):
        return self.location.line

    def reset(self):
        # -- RESET: Captured output data
        self.captured.reset()
        # -- RESET: ERROR CONTEXT INFO
        self.exception = None
        self.exc_traceback = None
        self.error_message = None

    def store_exception_context(self, exception):
        self.exception = exception
        self.exc_traceback = sys.exc_info()[2]

    def __hash__(self):
        # -- NEEDED-FOR: PYTHON3
        # return id((self.keyword, self.name))
        return id(self)

    def __eq__(self, other):
        # -- PYTHON3 SUPPORT, ORDERABLE:
        # NOTE: Ignore potential FileLocation differences.
        return (self.keyword, self.name) == (other.keyword, other.name)

    def __lt__(self, other):
        # -- PYTHON3 SUPPORT, ORDERABLE:
        # NOTE: Ignore potential FileLocation differences.
        return (self.keyword, self.name) < (other.keyword, other.name)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        # -- SEE ALSO: python2.7, functools.total_ordering
        # return not other < self
        return other >= self

    def __gt__(self, other):
        # -- SEE ALSO: python2.7, functools.total_ordering
        assert isinstance(other, BasicStatement)
        return other < self

    def __ge__(self, other):
        # -- SEE ALSO: python2.7, functools.total_ordering
        # OR: return self >= other
        return not self < other     # pylint: disable=unneeded-not

    # def __cmp__(self, other):
    #     # -- NOTE: Ignore potential FileLocation differences.
    #     return cmp((self.keyword, self.name), (other.keyword, other.name))


class TagStatement(BasicStatement):

    def __init__(self, filename, line, keyword, name, tags):
        if tags is None:
            tags = []
        super().__init__(filename, line, keyword, name)
        self.tags = tags

    def should_run_with_tags(self, tag_expression):
        """Determines if statement should run when the tag expression is used.

        :param tag_expression:  Runner/config environment tags to use.
        :return: True, if examples should run. False, otherwise (skip it).
        """
        return tag_expression.check(self.tags)


class TagAndStatusStatement(BasicStatement):
    # final_status = ('passed', 'failed', 'skipped')
    final_status = (Status.passed, Status.failed, Status.skipped)

    def __init__(self, filename, line, keyword, name, tags):
        super().__init__(filename, line, keyword, name)
        self.tags = tags
        self.should_skip = False
        self.skip_reason = None
        self._cached_status = Status.untested

    def should_run_with_tags(self, tag_expression):
        """Determines if statement should run when the tag expression is used.

        :param tag_expression:  Runner/config environment tags to use.
        :return: True, if examples should run. False, otherwise (skip it).
        """
        return tag_expression.check(self.tags)

    @property
    def status(self):
        if self._cached_status not in self.final_status:
            # -- RECOMPUTE: As long as final status is not reached.
            self._cached_status = self.compute_status()
        return self._cached_status

    def set_status(self, value):
        if isinstance(value, str):
            value = Status.from_name(value)
        self._cached_status = value

    def clear_status(self):
        self._cached_status = Status.untested

    def reset(self):
        self.should_skip = False
        self.skip_reason = None
        self.clear_status()

    def compute_status(self):
        raise NotImplementedError


class Replayable:
    type = None

    def replay(self, formatter):
        getattr(formatter, self.type)(self)


# -----------------------------------------------------------------------------
# UTILITY FUNCTIONS:
# -----------------------------------------------------------------------------
def unwrap_function(func, max_depth=10):
    """Unwraps a function that is wrapped with :func:`functools.partial()`"""
    iteration = 0
    wrapped = getattr(func, "__wrapped__", None)
    while wrapped and iteration < max_depth:
        func = wrapped
        wrapped = getattr(func, "__wrapped__", None)
        iteration += 1
    return func
